import logging
from collections import deque
from pprint import pprint
from typing import Any, Dict, Tuple

import pandas as pd
import pytest
import yaml

from ludwig.api import LudwigModel
from ludwig.config_sampling.explore_schema import combine_configs, ConfigOption, explore_properties
from ludwig.config_validation.validation import get_schema
from ludwig.types import ModelConfigDict

from .configs import (
    COMBINER_TYPE_TO_COMBINE_FN_MAP,
    ECD_CONFIG_SECTION_TO_CONFIG,
    FEATURE_TYPE_TO_CONFIG_FOR_DECODER_LOSS,
    FEATURE_TYPE_TO_CONFIG_FOR_ENCODER_PREPROCESSING,
)


def defaults_config_generator(
    feature_type: str, allow_list: str, static_schema: Dict[str, Any] = None
) -> Tuple[ModelConfigDict, pd.DataFrame]:
    """Generate combinatorial configs for the defaults section of the Ludwig config.

    Args:
        feature_type: feature type to explore.
        allow_list: top-level parameter of the defaults sections that should be included.
    """
    assert isinstance(allow_list, str)
    assert allow_list in {"preprocessing", "encoder", "decoder", "loss"}

    schema = get_schema() if not static_schema else static_schema
    properties = schema["properties"]["defaults"]["properties"][feature_type]["properties"]
    raw_entry = deque([ConfigOption(dict(), False)])
    explored = explore_properties(
        properties, parent_parameter_path=f"defaults.{feature_type}", dq=raw_entry, allow_list=[allow_list]
    )

    if allow_list in ["preprocessing", "encoder"]:
        config = FEATURE_TYPE_TO_CONFIG_FOR_ENCODER_PREPROCESSING[feature_type]
        config = yaml.safe_load(config)
    else:  # decoder and loss
        config = FEATURE_TYPE_TO_CONFIG_FOR_DECODER_LOSS[feature_type]
        config = yaml.safe_load(config)

    config["model_type"] = "ecd"
    config["trainer"] = {"train_steps": 1}

    combined_configs = combine_configs(explored, config)
    logging.info(f"Generated {len(combined_configs)} for {feature_type} {allow_list} combinatorial tests.")

    for config, dataset in combined_configs:
        yield config, dataset


def ecd_trainer_config_generator(static_schema: Dict[str, Any] = None) -> Tuple[ModelConfigDict, pd.DataFrame]:
    """Generate combinatorial configs for the ECD trainer section of the Ludwig config."""
    schema = get_schema() if not static_schema else static_schema
    properties = schema["properties"]

    raw_entry = deque([ConfigOption(dict(), False)])
    explored = explore_properties(properties, parent_parameter_path="", dq=raw_entry, allow_list=["trainer"])
    config = ECD_CONFIG_SECTION_TO_CONFIG["trainer"]
    config = yaml.safe_load(config)
    config["model_type"] = "ecd"
    config["trainer"] = {"train_steps": 1}

    combined_configs = combine_configs(explored, config)

    # HACK(Arnav): Remove configs that have LARS, LAMB or Lion optimizers, or Paged or 8-bit optimizers.
    # This is because they require GPUs.
    filtered_configs = []

    for config, dataset in combined_configs:
        optimizer_type = config.get("trainer", {}).get("optimizer", "").get("type", "")

        if optimizer_type not in {"lars", "lamb", "lion"} and not (
            "paged" in optimizer_type or "8bit" in optimizer_type
        ):
            filtered_configs.append((config, dataset))

    # Replace combined_configs with the filtered_configs
    combined_configs = filtered_configs

    logging.info(f"Generated {len(combined_configs)} for ECD trainer combinatorial tests.")

    for config, dataset in combined_configs:
        yield config, dataset


def combiner_config_generator(
    combiner_type: str, static_schema: Dict[str, Any] = None
) -> Tuple[ModelConfigDict, pd.DataFrame]:
    """Generate combinatorial configs for the combiner section of the Ludwig config.

    Args:
        combiner_type: combiner type to explore.
    """
    schema = get_schema() if not static_schema else static_schema
    properties = schema["properties"]

    raw_entry = deque([ConfigOption(dict(), False)])
    explored = explore_properties(properties, parent_parameter_path="", dq=raw_entry, allow_list=["combiner"])
    config = ECD_CONFIG_SECTION_TO_CONFIG[combiner_type]
    config = yaml.safe_load(config)
    config["model_type"] = "ecd"
    config["trainer"] = {"train_steps": 1}

    combine_configs_fn = COMBINER_TYPE_TO_COMBINE_FN_MAP[combiner_type]

    combined_configs = combine_configs_fn(explored, config)
    combined_configs = [
        (config, dataset) for config, dataset in combined_configs if config["combiner"]["type"] == combiner_type
    ]
    logging.info(f"Generated {len(combined_configs)} for {combiner_type} combiner combinatorial tests.")

    for config, dataset in combined_configs:
        yield config, dataset


def train_and_evaluate(config: ModelConfigDict, dataset: pd.DataFrame):
    """Trains and evaluates a model with the given config.

    Args:
        config: valid Ludwig config.
        dataset: Ludwig dataset name to train on.
    """
    # adding print statements to be captured in pytest stdout and help debug tests.
    print("Config used (trained on synthetic data)")
    pprint(config)
    model = LudwigModel(config=config, callbacks=None, logging_level=logging.ERROR)
    model.train(dataset=dataset)
    model.evaluate(dataset=dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", combiner_config_generator("sequence_concat"))
def test_ecd_sequence_concat_combiner(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", combiner_config_generator("sequence"))
def test_ecd_sequence_combiner(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", combiner_config_generator("comparator"))
def test_ecd_comparator_combiner(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", combiner_config_generator("concat"))
def test_ecd_concat_combiner(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", combiner_config_generator("project_aggregate"))
def test_ecd_project_aggregate_combiner(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", combiner_config_generator("tabnet"))
def test_ecd_tabnet_combiner(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", combiner_config_generator("tabtransformer"))
def test_ecd_tabtransformer_combiner(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", combiner_config_generator("transformer"))
def test_ecd_transformer_combiner(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", ecd_trainer_config_generator())
def test_ecd_trainer(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("number", "encoder"))
def test_number_encoder_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("number", "decoder"))
def test_number_decoder_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("number", "loss"))
def test_number_encoder_loss(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("number", "preprocessing"))
def test_number_preprocessing_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("category", "encoder"))
def test_category_encoder_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("category", "decoder"))
def test_category_decoder_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("category", "loss"))
def test_category_loss_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("category", "preprocessing"))
def test_category_preprocessing_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("binary", "encoder"))
def test_binary_encoder_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("binary", "decoder"))
def test_binary_decoder_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("binary", "loss"))
def test_binary_loss_defaults(config, dataset):
    train_and_evaluate(config, dataset)


@pytest.mark.combinatorial
@pytest.mark.parametrize("config,dataset", defaults_config_generator("binary", "preprocessing"))
def test_binary_preprocessing_defaults(config, dataset):
    train_and_evaluate(config, dataset)


# @pytest.mark.combinatorial
# @pytest.mark.parametrize("config,dataset", defaults_config_generator("text", "preprocessing"))
# def test_text_preprocessing_defaults(config, dataset):
#     train_and_evaluate(config, dataset)


# @pytest.mark.combinatorial
# @pytest.mark.parametrize("config,dataset", defaults_config_generator("text", "encoder"))
# def test_text_encoder_defaults(config):
#     train_and_evaluate(config, dataset)



from copy import deepcopy

import pandas as pd
import torch

from ludwig.api import LudwigModel
from ludwig.constants import COLUMN, INPUT_FEATURES, PREPROCESSING, SPLIT
from ludwig.data.split import get_splitter
from ludwig.features.base_feature import BaseFeature


def filter_cols(df, cols):
    cols = {c.lower() for c in cols}
    retain_cols = [c for c in df.columns if c.lower() in cols]
    return df[retain_cols]


def prepare_data(model: LudwigModel, inputs_df: pd.DataFrame, sample_df: pd.DataFrame, target: str):
    config = model.config
    feature_cols = [feature[COLUMN] for feature in config[INPUT_FEATURES]]
    if SPLIT in config.get(PREPROCESSING, {}):
        # Keep columns required for Ludwig preprocessing
        splitter = get_splitter(**config[PREPROCESSING][SPLIT])
        feature_cols += splitter.required_columns
    target_feature_name = get_feature_name(model, target)

    inputs_df = filter_cols(inputs_df, feature_cols)
    if sample_df is not None:
        sample_df = filter_cols(sample_df, feature_cols)

    return inputs_df, sample_df, feature_cols, target_feature_name


def get_pred_col(preds, target):
    t = target.lower()
    for c in preds.keys():
        if c.lower() == t:
            if "probabilities" in preds[c]:
                return preds[c]["probabilities"]
            else:
                return preds[c]["predictions"]
    raise ValueError(f"Unable to find target column {t} in {preds.keys()}")


def get_feature_name(model: LudwigModel, target: str) -> str:
    t = target.lower()
    for c in model.training_set_metadata.keys():
        if c.lower() == t:
            return c
    raise ValueError(f"Unable to find target column {t} in {model.training_set_metadata.keys()}")


def get_absolute_module_key_from_submodule(module: torch.nn.Module, submodule: torch.nn.Module):
    """Get the absolute module key for each param in the target layer.

    Assumes that the keys in the submodule are relative to the module.

    We find the params from the submodule in the module by comparing the data
    pointers, since the data returned by named_parameters is by reference.
    More information on checking if tensors point to the same place in storage can be found here:
    https://discuss.pytorch.org/t/any-way-to-check-if-two-tensors-have-the-same-base/44310/2
    """
    absolute_keys = []
    for module_key, module_param in module.named_parameters():
        for _, submodule_param in submodule.named_parameters():
            if submodule_param.data_ptr() == module_param.data_ptr():
                absolute_keys.append(module_key)
                break
    return absolute_keys


def replace_layer_with_copy(feat: BaseFeature, target_layer: torch.nn.Module):
    """Replaces a layer in a feature with a copy of the layer in-place.

    This is useful in a tied weights scenario, where a single encoder may be used by multiple features. If we leave
    as-is, Captum complains about the resulting computation graph. The solution is to create an identical
    (deep) copy of the layer fed into Captum: https://github.com/pytorch/captum/issues/794#issuecomment-1093021638

    This is safe to do during the explain step because we are essentially running inference, and no model artifacts are
    being saved during the explain step.

    TODO(geoffrey): if a user ever wants to train immediately after explain (i.e. w/o loading weights from the disk),
    we might want to implement this as a context so that we can restore the original encoder object at the end.
    Will defer this implementation for now because that scenario seems unlikely.

    At a high-level the approach is the following:
    1. Create a deep-copy of the entire encoder object and set it as the feature's encoder object
    2. Replace the tensors in the copied encoder object with the tensors from the original encoder object, except for
         the tensors in the target layer. We want to explain these tensors, so we want to keep them as deep copies.

    This approach ensures that at most 2 copies of the encoder object are in memory at any given time.
    """
    with torch.no_grad():
        # Get the original encoder object and a mapping from param names to the params themselves.
        orig_encoder_obj = feat.encoder_obj
        orig_encoder_obj_state_dict = orig_encoder_obj.state_dict()

        # Deep copy the original encoder object and set the copy as this feature's encoder object.
        copy_encoder_obj = deepcopy(orig_encoder_obj)
        feat.encoder_obj = copy_encoder_obj

        # We have to get the absolute module key in order to do string matching because the target_layer keys are
        # relative to itself. If we were to leave it as-is and attempt to suffix match, we may get duplicates for
        # common layers i.e. "LayerNorm.weight" and "LayerNorm.bias". Getting the absolute module key ensures we
        # use values like "transformer.module.embedding.LayerNorm.weight" instead.
        keys_to_keep_copy = get_absolute_module_key_from_submodule(orig_encoder_obj, target_layer)

        # Get the tensors to keep from the copied encoder object. These are the tensors in the target layer.
        for key, param in copy_encoder_obj.named_parameters():
            if key not in keys_to_keep_copy:
                param.data = orig_encoder_obj_state_dict[key].data



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import SEMANTIC
from ludwig.error import ConfigValidationError
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import LLM_METADATA
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class RetrievalConfig(schema_utils.BaseMarshmallowConfig):
    """This Dataclass is a schema for the nested retrieval config under prompt."""

    def __post_init__(self):
        # TODO: have a dynamically loaded schema based on the selection of the type param
        # https://github.com/ludwig-ai/ludwig/pull/3351#discussion_r1181910954
        # Ensure k is non-zero if we're using a retrieval strategy
        if self.type is not None and self.k == 0:
            self.k = 1

        if self.type is None and self.k != 0:
            raise ConfigValidationError("k must be 0 if retrieval type is None.")
        elif self.type is not None and self.k <= 0:
            raise ConfigValidationError("k must be greater than 0 if retrieval type is not None.")

        if self.type is None and self.model_name is not None:
            raise ConfigValidationError("model_name must be None if retrieval type is None.")
        elif self.type == SEMANTIC and self.model_name is None:
            raise ConfigValidationError(f"model_name must not be None if retrieval type is '{SEMANTIC}'.")

    type: str = schema_utils.String(
        default=None,
        allow_none=True,
        description=(
            "The type of retrieval to use for the prompt. If `None`, then no retrieval is used, and the task "
            "is framed as a zero-shot learning problem. If not `None` (e.g. either 'random' or 'semantic'), then "
            "samples are retrieved from an index of the training set and used to augment the input to the model "
            "in a few-shot learning setting."
        ),
        parameter_metadata=LLM_METADATA["prompt"]["retrieval"]["type"],
    )

    index_name: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="The name of the index to use for the prompt. Indices are stored in the ludwig cache by default.",
        parameter_metadata=LLM_METADATA["prompt"]["retrieval"]["index_name"],
    )

    model_name: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="The model used to generate the embeddings used to retrieve samples to inject in the prompt.",
        parameter_metadata=LLM_METADATA["prompt"]["retrieval"]["model_name"],
    )

    k: int = schema_utils.NonNegativeInteger(
        default=0,
        description="The number of samples to retrieve.",
        parameter_metadata=LLM_METADATA["prompt"]["retrieval"]["k"],
    )


@DeveloperAPI
class RetrievalConfigField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(RetrievalConfig)

    def _jsonschema_type_mapping(self):
        return schema_utils.unload_jsonschema_from_marshmallow_class(RetrievalConfig, title="Retrieval")


@DeveloperAPI
@ludwig_dataclass
class PromptConfig(schema_utils.BaseMarshmallowConfig):
    """This Dataclass is a schema for the nested prompt config under preprocessing."""

    template: str = schema_utils.String(
        default=None,
        allow_none=True,
        description=(
            "The template to use for the prompt. Must contain at least one of the columns from the input dataset "
            "or `__sample__` as a variable surrounded in curly brackets {} to indicate where to insert the "
            "current feature. Multiple columns can be inserted, e.g.: `The {color} {animal} jumped over "
            "the {size} {object}`, where every term in curly brackets is a column in the dataset. If a `task` "
            "is specified, then the template must also contain the `__task__` variable. If `retrieval` is specified, "
            "then the template must also contain the `__context__` variable. If no template is provided, then a "
            "default will be used based on the retrieval settings, and a task must be set in the config."
        ),
        parameter_metadata=LLM_METADATA["prompt"]["template"],
    )

    task: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="The task to use for the prompt. Required if `template` is not set.",
        parameter_metadata=LLM_METADATA["prompt"]["task"],
    )

    retrieval: RetrievalConfig = RetrievalConfigField().get_default_field()


@DeveloperAPI
class PromptConfigField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(PromptConfig)

    def _jsonschema_type_mapping(self):
        return schema_utils.unload_jsonschema_from_marshmallow_class(PromptConfig)



from typing import Dict, List, Optional, Tuple, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import LLM_METADATA


@DeveloperAPI
@schema_utils.ludwig_dataclass
class LLMGenerationConfig(schema_utils.BaseMarshmallowConfig):
    """Parameters for LLM Generation Config.

    Should match the parameters in
    https://huggingface.co/docs/transformers/v4.28.0/en/main_classes/text_generation#transformers.GenerationConfig
    """

    # Parameters that control the length of the output

    max_new_tokens: Optional[int] = schema_utils.PositiveInteger(
        default=32,
        allow_none=True,
        description="The maximum number of new tokens to generate, ignoring the number of tokens in the input prompt.",
        parameter_metadata=LLM_METADATA["generation"]["max_new_tokens"],
    )

    min_new_tokens: Optional[int] = schema_utils.NonNegativeInteger(
        default=None,
        allow_none=True,
        description="The minimum number of new tokens to generate, ignoring the number of tokens in the input prompt.",
        parameter_metadata=LLM_METADATA["generation"]["min_new_tokens"],
    )

    max_length: int = schema_utils.PositiveInteger(
        default=32,
        allow_none=True,
        description="The maximum length the generated tokens can have. Corresponds to the length of the input prompt "
        "+ max_new_tokens. Its effect is overridden by max_new_tokens, if also set.",
        parameter_metadata=LLM_METADATA["generation"]["max_length"],
    )

    min_length: int = schema_utils.NonNegativeInteger(
        default=0,
        allow_none=True,
        description="The minimum length of the sequence to be generated. Corresponds to the length of the "
        "input prompt + min_new_tokens. Its effect is overridden by min_new_tokens, if also set.",
        parameter_metadata=LLM_METADATA["generation"]["min_length"],
    )

    early_stopping: Optional[Union[bool, str]] = schema_utils.Boolean(
        default=False,
        description="Controls the stopping condition for beam-based methods, like beam-search. It accepts the following"
        " values: True, where the generation stops as soon as there are num_beams complete candidates; False, where an "
        "heuristic is applied and the generation stops when is it very unlikely to find better candidates; `never`, "
        "where the beam search procedure only stops when there cannot be better candidates (canonical beam search "
        "algorithm)",
    )

    max_time: Optional[float] = schema_utils.FloatRange(
        default=None,
        min=None,
        max=None,
        allow_none=True,
        description="The maximum amount of time you allow the computation to run for in seconds. generation will still"
        " finish the current pass after allocated time has been passed. ",
    )

    # Parameters that control the generation strategy used

    do_sample: Optional[bool] = schema_utils.Boolean(
        default=True,
        description="Whether or not to use sampling ; use greedy decoding otherwise.",
        parameter_metadata=LLM_METADATA["generation"]["do_sample"],
    )

    num_beams: Optional[int] = schema_utils.PositiveInteger(
        default=1,
        allow_none=True,
        description="Number of beams for beam search. 1 means no beam search and is the default value."
        " The beam search strategy generates the translation word by word from left-to-right while keeping a fixed"
        " number (beam) of active candidates at each time step during token generation. By increasing the beam size,"
        " the translation performance can increase at the expense of significantly reducing the decoder speed.",
        parameter_metadata=LLM_METADATA["generation"]["num_beams"],
    )

    num_beam_groups: Optional[int] = schema_utils.PositiveInteger(
        default=1,
        allow_none=True,
        description="Number of groups to divide num_beams into in order to ensure diversity among different groups of "
        "beams. 1 means no group beam search.",
    )

    penalty_alpha: Optional[float] = schema_utils.NonNegativeFloat(
        default=None,
        allow_none=True,
        description="The values balance the model confidence and the degeneration penalty in contrastive "
        " search decoding.",
    )

    use_cache: Optional[bool] = schema_utils.Boolean(
        default=True,
        description="Whether or not the model should use the past last key/values attentions (if applicable to the "
        "model) to speed up decoding.",
        parameter_metadata=LLM_METADATA["generation"]["use_cache"],
    )

    # Parameters for manipulation of the model output logits

    temperature: Optional[float] = schema_utils.NonNegativeFloat(
        default=0.1,
        allow_none=True,
        description="Temperature is used to control the randomness of predictions."
        " A high temperature value (closer to 1) makes the output more diverse and random, while a lower temperature"
        " (closer to 0) makes the model's responses more deterministic and focused on the most likely outcome."
        " In other words, temperature adjusts the probability distribution from which the model picks the next token.",
        parameter_metadata=LLM_METADATA["generation"]["temperature"],
    )

    top_k: Optional[int] = schema_utils.PositiveInteger(
        default=50,
        allow_none=True,
        description="The number of highest probability vocabulary tokens to keep for top-k-filtering.",
        parameter_metadata=LLM_METADATA["generation"]["top_k"],
    )

    top_p: Optional[float] = schema_utils.FloatRange(
        default=1.0,
        min=0.0,
        max=1.0,
        allow_none=True,
        description="If set to float < 1, only the most probable tokens with probabilities that add up to "
        "top_p or higher are kept for generation.",
        parameter_metadata=LLM_METADATA["generation"]["top_p"],
    )

    typical_p: Optional[float] = schema_utils.FloatRange(
        default=1.0,
        min=0.0,
        max=1.0,
        allow_none=True,
        description="Local typicality measures how similar the conditional probability of predicting a target token "
        "next is to the expected conditional probability of predicting a random token next, given the partial text "
        "already generated. If set to float < 1, the smallest set of the most locally typical tokens with "
        "probabilities that add up to typical_p or higher are kept for generation.",
    )

    epsilon_cutoff: Optional[float] = schema_utils.FloatRange(
        default=0.0,
        min=0.0,
        max=1.0,
        allow_none=True,
        description="If set to float strictly between 0 and 1, only tokens with a conditional probability greater "
        "than epsilon_cutoff will be sampled. In the paper, suggested values range from 3e-4 to 9e-4, depending on the"
        " size of the model.",
    )

    eta_cutoff: Optional[float] = schema_utils.FloatRange(
        default=0.0,
        min=0.0,
        max=1.0,
        allow_none=True,
        description="Eta sampling is a hybrid of locally typical sampling and epsilon sampling. If set to float "
        "strictly between 0 and 1, a token is only considered if it is greater than either eta_cutoff or "
        "sqrt(eta_cutoff) * exp(-entropy(softmax(next_token_logits))). The latter term is intuitively the expected next"
        " token probability, scaled by sqrt(eta_cutoff). In the paper, suggested values range from 3e-4 to 2e-3, "
        "depending on the size of the model.",
    )

    diversity_penalty: Optional[float] = schema_utils.NonNegativeFloat(
        default=0.0,
        allow_none=True,
        description="The value used to control the diversity of the generated text. The higher the value, the more "
        "diverse the text will be. If set to 0, no diversity is enforced."
        "This value is subtracted from a beam(s) score if it generates a token same as any beam from other group at a"
        "particular time. Note that diversity_penalty is only effective if group beam search is enabled.",
    )

    repetition_penalty: Optional[float] = schema_utils.NonNegativeFloat(
        default=1.0,
        allow_none=True,
        description="The parameter for repetition penalty. 1.0 means no penalty. "
        "See [this paper](https://arxiv.org/pdf/1909.05858.pdf) for more details.",
    )

    encoder_repetition_penalty: Optional[float] = schema_utils.NonNegativeFloat(
        default=1.0,
        allow_none=True,
        description="The paramater for encoder_repetition_penalty. An exponential penalty on sequences that are not"
        " in the original input. 1.0 means no penalty.",
    )

    length_penalty: Optional[float] = schema_utils.Float(
        default=1.0,
        allow_none=True,
        description="Exponential penalty to the length that is used with beam-based generation. It is applied as an "
        "exponent to the sequence length, which in turn is used to divide the score of the sequence. Since the score is"
        " the log likelihood of the sequence (i.e. negative), length_penalty > 0.0 promotes longer sequences, while "
        "length_penalty < 0.0 encourages shorter sequences.",
    )

    no_repeat_ngram_size: Optional[int] = schema_utils.NonNegativeInteger(
        default=0,
        allow_none=True,
        description="If set to int > 0, all ngrams of that size can only occur once.",
    )

    bad_words_ids: Optional[List[List[int]]] = schema_utils.List(
        default=None,
        allow_none=True,
        description="List of token ids that are not allowed to be generated. In order to get the tokens of the words "
        "that should not appear in the generated text, use tokenizer(bad_word, add_prefix_space=True).input_ids.",
    )

    force_words_ids: Optional[List[List[int]]] = schema_utils.List(
        default=None,
        allow_none=True,
        description="List of token ids that are forced to be generated by the model. In order to get the tokens of the"
        " words that should appear in the generated text, use tokenizer(force_word, add_prefix_space=True).input_ids.",
    )

    renormalize_logits: Optional[bool] = schema_utils.Boolean(
        default=False,
        description="Whether to renormalize the logits after temperature and top_k/top_p filtering.",
    )

    # TODO(This needs to be defined based on the Constraint class)
    # constraints:

    forced_bos_token_id: Optional[int] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description="The id of the token to force as the first generated token after the decoder_start_token_id."
        "Useful for multilingual models like mBART where the first generated token needs to be the target language"
        "token.",
    )

    forced_eos_token_id: Optional[Union[int, List[int]]] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description="The id of the token to force as the last generated token when max_length is reached. Optionally, "
        "use a list to set multiple end-of-sequence tokens.",
    )

    remove_invalid_values: Optional[bool] = schema_utils.Boolean(
        default=False,
        description="Whether to remove possible nan and inf outputs of the model to prevent the generation method to "
        "crash. Note that using remove_invalid_values can slow down generation.",
    )

    exponential_decay_length_penalty: Optional[Tuple[int, float]] = schema_utils.FloatRange(
        default=None,
        min=0.0,
        max=1.0,
        allow_none=True,
        description="This Tuple adds an exponentially increasing length penalty, after a certain amount of tokens have "
        "been generated. The tuple shall consist of: (start_index, decay_factor) where start_index indicates where "
        "penalty starts and decay_factor represents the factor of exponential decay",
    )

    suppress_tokens: Optional[List[int]] = schema_utils.List(
        list_type=int,
        default=None,
        allow_none=True,
        description="A list of tokens that will be suppressed at generation. The SupressTokens logit processor will set"
        " their log probs to -inf so that they are not sampled.",
    )

    begin_suppress_tokens: Optional[List[int]] = schema_utils.List(
        list_type=int,
        default=None,
        allow_none=True,
        description="A list of tokens that will be suppressed at the beginning of the generation. The "
        "SupressBeginTokens logit processor will set their log probs to -inf so that they are not sampled.",
    )

    forced_decoder_ids: Optional[List[List[int]]] = schema_utils.List(
        default=None,
        allow_none=True,
        description="A list of forced decoder ids. The ForcedDecoderIds logit processor will set the log probs of all "
        "tokens that are not in the list to -inf so that they are not sampled.",
    )

    sequence_bias: Optional[Dict[Tuple[int], float]] = schema_utils.Dict(
        default=None,
        allow_none=True,
        description="A dictionary of token ids to bias the generation towards. The SequenceBias logit processor will "
        "add the bias to the log probs of the tokens in the dictionary. Positive biases increase the odds of the "
        "sequence being selected, while negative biases do the opposite. ",
    )

    guidance_scale: Optional[float] = schema_utils.FloatRange(
        default=None,
        min=0.0,
        allow_none=True,
        description="The guidance scale for classifier free guidance (CFG). CFG is enabled by setting guidance_scale >"
        " 1. Higher guidance scale encourages the model to generate samples that are more closely linked to the input"
        " prompt, usually at the expense of poorer quality.",
    )

    # Special tokens that can be used at generation time

    pad_token_id: Optional[int] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description="The id of the padding token. If not set, the padding token id of the tokenizer is used.",
    )

    bos_token_id: Optional[int] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description="The id of the beginning of sentence token. If not set, the bos token id of the tokenizer is used.",
    )

    eos_token_id: Optional[Union[int, List[int]]] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description="The id of the end of sentence token. If not set, the eos token id of the tokenizer is used.",
    )


@DeveloperAPI
class LLMGenerationConfigField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(LLMGenerationConfig)

    def _jsonschema_type_mapping(self):
        return schema_utils.unload_jsonschema_from_marshmallow_class(LLMGenerationConfig)



import contextlib
import logging
import os
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer, GenerationConfig, PreTrainedModel

from ludwig.constants import IGNORE_INDEX_TOKEN_ID, LOGITS, MODEL_LLM, PREDICTIONS, TEXT
from ludwig.features.base_feature import ModuleWrapper, OutputFeature
from ludwig.features.feature_utils import LudwigFeatureDict
from ludwig.features.text_feature import TextOutputFeature
from ludwig.globals import MODEL_WEIGHTS_FILE_NAME
from ludwig.models.base import BaseModel
from ludwig.schema.features.base import BaseOutputFeatureConfig, FeatureCollection
from ludwig.schema.model_types.llm import LLMModelConfig
from ludwig.utils.augmentation_utils import AugmentationPipelines
from ludwig.utils.data_utils import clear_data_cache
from ludwig.utils.llm_utils import (
    add_left_padding,
    generate_merged_ids,
    pad_target_tensor_for_fine_tuning,
    realign_target_and_prediction_tensors_for_inference,
    remove_left_padding,
    set_pad_token,
)
from ludwig.utils.logging_utils import log_once
from ludwig.utils.output_feature_utils import set_output_feature_tensor
from ludwig.utils.torch_utils import reg_loss

logger = logging.getLogger(__name__)


class DictWrapper:
    """Wrapper for a LudwigFeatureDict module that allows for iteration over keys.

    The purpose of this class is to avoid exposing input and output features as modules of the LLM. This is because we
    only wish to train the underlying model, and having these additional modules can confuse systems like DeepSpeed.
    """

    def __init__(self, obj: LudwigFeatureDict):
        self.obj = obj

    def get(self, key) -> torch.nn.Module:
        return self.obj.get(key)

    def set(self, key: str, module: torch.nn.Module) -> None:
        self.obj.set(key, module)

    def __len__(self) -> int:
        return len(self.obj)

    def __next__(self) -> None:
        return next(iter(self.obj))

    def __iter__(self) -> None:
        return iter(self.obj.keys())

    def keys(self) -> List[str]:
        return self.obj.keys()

    def values(self) -> List[torch.nn.Module]:
        return self.obj.values()

    def items(self) -> List[Tuple[str, torch.nn.Module]]:
        return self.obj.items()

    def update(self, modules: Dict[str, torch.nn.Module]) -> None:
        self.obj.update(modules)


def load_pretrained_from_config(
    config_obj: LLMModelConfig,
    model_config: Optional[AutoConfig] = None,
    weights_save_path: Optional[str] = None,
) -> PreTrainedModel:
    load_kwargs = {}
    if config_obj.quantization:
        # Apply quanitzation configuration at model load time
        load_kwargs["torch_dtype"] = getattr(torch, config_obj.quantization.bnb_4bit_compute_dtype)
        load_kwargs["quantization_config"] = config_obj.quantization.to_bitsandbytes()
        load_kwargs["device_map"] = "auto"

    if config_obj.model_parameters:
        # Add any model specific parameters to the load kwargs
        for param_name, param_value in config_obj.model_parameters.to_dict().items():
            # Not all parameters are supported by all models, so we only add the parameter to the load kwargs
            # if it is supported by the model.
            if param_value is None:
                continue

            if hasattr(model_config, param_name):
                load_kwargs[param_name] = param_value
            else:
                logger.warning(f"Parameter {param_name} is not supported by {config_obj.base_model}. Skipping.")

    logger.info("Loading large language model...")
    pretrained_model_name_or_path = weights_save_path or config_obj.base_model
    model: PreTrainedModel = AutoModelForCausalLM.from_pretrained(pretrained_model_name_or_path, **load_kwargs)
    return model


class LLM(BaseModel):
    @staticmethod
    def type() -> str:
        return MODEL_LLM

    def __init__(
        self,
        config_obj: LLMModelConfig,
        random_seed=None,
        device=None,
        **_kwargs,
    ):
        super().__init__(random_seed=random_seed)

        self.config_obj = config_obj
        self._random_seed = random_seed

        self.model_name = self.config_obj.base_model
        self.model_config = AutoConfig.from_pretrained(self.config_obj.base_model)

        self.model = load_pretrained_from_config(self.config_obj, model_config=self.model_config)
        self.curr_device = next(self.model.parameters()).device
        logger.info("Done.")

        # Determines the maximum length of the context (input + output tokens)
        if hasattr(self.model_config, "max_sequence_length"):
            self.context_len = self.model_config.max_sequence_length
        elif hasattr(self.model_config, "max_position_embeddings"):
            self.context_len = self.model_config.max_position_embeddings
        else:
            self.context_len = 2048

        # TODO(Arnav): This needs be more flexible to account for RoPE Scaling
        # When merging input IDs and target IDs for LLM fine-tuning, we want to make sure that the merged tensor is
        # not longer than the global maximum sequence length. This is provided in the preprocessing config. We never
        # want to exceed the maximum possible context length so we also check for that.
        if self.config_obj.preprocessing.global_max_sequence_length:
            global_max_sequence_length = self.config_obj.preprocessing.global_max_sequence_length
            self.global_max_sequence_length = (
                global_max_sequence_length if global_max_sequence_length <= self.context_len else self.context_len
            )
        else:
            self.global_max_sequence_length = self.context_len

        # Initialize tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(self.config_obj.base_model)
        set_pad_token(self.tokenizer)

        self._set_generation_config(self.config_obj.generation.to_dict())

        # ================ Inputs ================
        try:
            self.input_features.update(self.build_inputs(input_feature_configs=self.config_obj.input_features))
        except KeyError as e:
            raise KeyError(
                f"An input feature has a name that conflicts with a class attribute of torch's ModuleDict: {e}"
            )

        # This is used to store the model inputs during the forward pass when fine-tuning LLMs. This allows us to have
        # access to the joint model inputs (input_ids and target_ids) when computing metrics. In particular, the target
        # ids are needed to correctly compute next token softmax cross entropy loss.
        self.model_inputs = None

        # ================ Outputs ================
        self.output_feature_type = self.config_obj.output_features[0].type

        self.output_features.update(
            self.build_outputs(
                output_feature_configs=self.config_obj.output_features,
                # Set the input size to the model vocab size instead of the tokenizer vocab size
                # because the model has additional "head" layers that are used to predict the next
                # token in the sequence. These head layers can add additional dimensions to the
                # logits tensor, beyond the vocab_size dimension.
                input_size=self.input_shape[-1] if self.output_feature_type == TEXT else self.model_config.vocab_size,
            )
        )

        # Extract the decoder object for the forward pass
        self._output_feature_decoder = ModuleWrapper(self.output_features.items()[0][1])

        self.attention_masks = None

        clear_data_cache()

    def create_feature_dict(self) -> DictWrapper:
        return DictWrapper(LudwigFeatureDict())

    @contextlib.contextmanager
    def use_generation_config(self, generation_config_dict: Optional[Dict[str, Any]] = None):
        """Sets the generation config for the model."""
        # Save the original generation config so that we can reset it if/when we change it when self.generation gets is
        # dynamically mutated during 1-off predict calls after fine-tuning.
        original_generation_config_dict = self.generation.to_dict()
        try:
            # no-op if generation_config is None
            if generation_config_dict is not None:
                # unwrap the original generation config, update it with the new generation config
                new_generation_config_dict = {**original_generation_config_dict, **generation_config_dict}
                self._set_generation_config(new_generation_config_dict)
            yield
        finally:
            self._set_generation_config(original_generation_config_dict)

    def _set_generation_config(self, new_generation_config_dict: Dict[str, Any]):
        self.generation = GenerationConfig(**new_generation_config_dict)
        # We need to manually set the pad_token_id to the tokenizer's pad_token_id for certain models like GPT and
        # CodeLlama to avoid getting an error. This workaround can be found here:
        # (https://github.com/huggingface/transformers/issues/25353#issuecomment-1669339754)
        self.generation.pad_token_id = self.tokenizer.pad_token_id
        self.max_new_tokens = self.generation.max_new_tokens
        # max input length value copied from FastChat
        # https://github.com/lm-sys/FastChat/blob/0e958b852a14f4bef5f0e9d7a5e7373477329cf2/fastchat/serve/inference.py#L183  # noqa E501
        self.max_input_length = self.context_len - self.max_new_tokens - 8

    @property
    def output_feature_decoder(self) -> OutputFeature:
        return self._output_feature_decoder.module

    def initialize_adapter(self):
        """If an adapter config is provided, we want to wrap the model with a PEFT model for fine-tuning."""
        if self.config_obj.adapter:
            if self.config_obj.trainer.type != "finetune" and not self.config_obj.adapter.pretrained_adapter_weights:
                raise ValueError(
                    "Adapter config was provided, but trainer type is not set to `finetune`. Either set the trainer to "
                    "`finetune` or remove the adapter config."
                )

            from peft import get_peft_model

            if self.config_obj.adapter.pretrained_adapter_weights:
                logger.info(f"Using pretrained adapter weights: {self.config_obj.adapter.pretrained_adapter_weights}")
                # If pretrained adapter weights are provided, we want to load them into the model
                from peft import MODEL_TYPE_TO_PEFT_MODEL_MAPPING, PeftConfig

                peft_config = PeftConfig.from_pretrained(self.config_obj.adapter.pretrained_adapter_weights)

                self.model = MODEL_TYPE_TO_PEFT_MODEL_MAPPING[peft_config.task_type].from_pretrained(
                    self.model, self.config_obj.adapter.pretrained_adapter_weights
                )
            else:
                # If no pretrained adapter is provided, we want to load untrained weights into the model
                from peft import TaskType

                peft_config = self.config_obj.adapter.to_config(
                    task_type=TaskType.CAUSAL_LM, tokenizer_name_or_path=self.model_name
                )

                self.model = get_peft_model(self.model, peft_config)

            logger.info("==================================================")
            logger.info("Trainable Parameter Summary For Fine-Tuning")
            logger.info(f"Fine-tuning with adapter: {self.config_obj.adapter.type}")
            self.model.print_trainable_parameters()
            logger.info("==================================================")

    def prepare_for_training(self):
        # TODO: this implementation will not work if resuming from a previous checkpoint. Need to fix this.
        if self.config_obj.quantization:
            self.prepare_for_quantized_training()
        self.initialize_adapter()

    def prepare_for_quantized_training(self):
        from peft import prepare_model_for_kbit_training

        self.model = prepare_model_for_kbit_training(self.model, use_gradient_checkpointing=False)

    def to_device(self, device):
        device = torch.device(device)

        if device.type == self.curr_device.type:
            log_once(f"Model already on device'{device}'.")
            return self
        else:
            log_once(f"Moving LLM from '{self.curr_device}' to '{device}'.")

        model_kwargs = {}
        num_gpus = torch.cuda.device_count()
        if device == torch.device("cuda") and num_gpus > 1:
            # TODO: make this configurable in the future. These parameters are from FastChat:
            # https://github.com/lm-sys/FastChat/blob/0e958b852a14f4bef5f0e9d7a5e7373477329cf2/fastchat/serve/inference.py#L90  # noqa
            # TODO: Wrap device_map="auto" in a try-except block since it may not be supported for all models (E.g. BertLMHead)  # noqa
            # We don't add quantization here (float16 or bfloat16) since we may not always want to quantize. We should
            # make quantization configurable in the future via the trainer config.
            model_kwargs.update(
                dict(
                    low_cpu_mem_usage=True,
                    device_map="auto",
                    max_memory={i: "13GiB" for i in range(num_gpus)},
                )
            )

            if self.config_obj.quantization:
                model_kwargs["quantization_config"] = self.config_obj.quantization.to_bitsandbytes()

            # we save and reload the weights to ensure that they can be sharded across the GPUs using `from_pretrained`
            with tempfile.TemporaryDirectory() as tmpdir:
                self.model.save_pretrained(tmpdir)

                if self.config_obj.adapter:
                    from peft import PeftModel

                    self.model = AutoModelForCausalLM.from_pretrained(
                        self.model_name,
                        **model_kwargs,
                    )
                    self.model = PeftModel.from_pretrained(
                        self.model,
                        tmpdir,
                        torch_dtype=torch.float16,
                    )
                else:
                    self.model = AutoModelForCausalLM.from_pretrained(
                        tmpdir,
                        **model_kwargs,
                    )

        else:
            self.model = self.model.to(device)

        self.curr_device = device
        return self

    @classmethod
    def build_outputs(
        cls, output_feature_configs: FeatureCollection[BaseOutputFeatureConfig], input_size: int
    ) -> Dict[str, OutputFeature]:
        """Builds and returns output feature."""
        # TODO: only single task currently
        if len(output_feature_configs) > 1:
            raise ValueError("The LLM model type only supports a single output feature.")

        output_feature_config = output_feature_configs[0]
        output_feature_config.input_size = input_size

        output_features = {}
        output_feature = cls.build_single_output(output_feature_config, output_features)
        output_features[output_feature_config.name] = output_feature

        return output_features

    def forward(
        self,
        inputs: Union[
            Dict[str, torch.Tensor], Dict[str, np.ndarray], Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]
        ],
        mask=None,
    ) -> Dict[str, torch.Tensor]:
        """Produces logits tensor for finetuning the model.

        Args:
            inputs: Inputs to the model. Can be a dictionary of input names to
                input tensors or a tuple of (inputs, targets) where inputs is
                a dictionary of input names to input tensors and targets is a
                dictionary of target names to target tensors.
            mask: A mask for the inputs.

        Returns:
            A dictionary of output {feature name}::{tensor_name} -> output tensor.
        """
        input_ids, target_ids = self._unpack_inputs(inputs)

        # Generate merged input_id, target_id pairs for the model, and create corresponding attention masks
        # We save them as class variables so that we can use them when realigning target and prediction tensors
        self.model_inputs, self.attention_masks = generate_merged_ids(
            input_ids, target_ids, self.tokenizer, self.global_max_sequence_length
        )

        # Wrap with flash attention backend for faster generation
        with torch.backends.cuda.sdp_kernel(enable_flash=True, enable_math=False, enable_mem_efficient=False) if (
            torch.cuda.is_available() and self.curr_device.type == "cuda"
        ) else contextlib.nullcontext():
            # TODO (jeffkinnison): Determine why the 8-bit `SCB` and `CB` matrices are deleted in the forward pass
            model_outputs = self.model(input_ids=self.model_inputs, attention_mask=self.attention_masks).get(LOGITS)

        if self.output_feature_type != TEXT:
            # Pass generated tokens through decoder after averaging the token probabilities
            # This is required for the classification head for the classifier decoder
            model_outputs = torch.mean(model_outputs, dim=1)

        if self.output_feature_type == TEXT:
            decoder_outputs = model_outputs
        else:
            decoder_outputs = self.output_feature_decoder.decoder_obj(model_outputs)

        # Set the output feature tensor to the decoder outputs (logits)
        outputs = {}
        of_name = self.config_obj.output_features[0].name
        set_output_feature_tensor(outputs, of_name, LOGITS, decoder_outputs)

        # Get predictions, probabilities and logits tensor from the output feature's predictions function
        outputs = self.output_features.get(of_name).predictions(outputs, of_name)

        # Cast to float32 for metric computation incase we're using deespeed with
        # reduced precision such as bfloat16.
        for prediction_key, prediction_tensor in outputs.items():
            if prediction_key != PREDICTIONS:
                # Skipping casting it to float32 since the predictions are tokens and they should be int64
                # (which is already the case)
                outputs[prediction_key] = prediction_tensor.type(torch.float32)

        return outputs

    def generate(
        self,
        inputs: Union[
            Dict[str, torch.Tensor], Dict[str, np.ndarray], Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]
        ],
        mask=None,
    ) -> Dict[str, torch.Tensor]:
        """Generates tokens using the model."""
        logger.info(f"For generating text, using: {self.generation}")
        input_ids, _ = self._unpack_inputs(inputs)

        with torch.no_grad():
            input_lengths = []
            sequences_list = []
            for input_ids_sample in input_ids:
                input_ids_sample_no_padding = remove_left_padding(input_ids_sample, self.tokenizer)
                logger.info(
                    "Decoded text inputs for the first example in batch: "
                    f"{self.tokenizer.decode(input_ids_sample_no_padding[0], skip_special_tokens=True)}"
                )

                if input_ids_sample_no_padding.shape[1] > self.max_input_length:
                    logger.warning(
                        f"Input length {input_ids_sample_no_padding.shape[1]} is "
                        f"greater than max input length {self.max_input_length}. Truncating."
                    )
                    input_ids_sample_no_padding = input_ids_sample_no_padding[:, -self.max_input_length :]  # noqa E203

                input_lengths.append(input_ids_sample_no_padding.shape[1])

                # Wrap with flash attention backend for faster generation
                with torch.backends.cuda.sdp_kernel(
                    enable_flash=True, enable_math=False, enable_mem_efficient=False
                ) if (torch.cuda.is_available() and self.curr_device.type == "cuda") else contextlib.nullcontext():
                    # Generate text using the model
                    model_outputs = self.model.generate(
                        input_ids=input_ids_sample_no_padding,
                        attention_mask=mask,
                        generation_config=self.generation,
                        return_dict_in_generate=True,
                        output_scores=True,
                    )
                    logger.info(
                        "Decoded generated output for the first example in batch: "
                        f"{self.tokenizer.batch_decode(model_outputs.sequences, skip_special_tokens=True)[0]}"
                    )

                sequences_list.append(model_outputs.sequences[0])

            # Extract the predictions, probabilities and logits from the model outputs
            # through the forward pass of the output feature
            outputs = self.output_feature_decoder.decoder_obj.forward(
                sequences_list,
                input_lengths,
                self.max_new_tokens,
            )

        return outputs

    def is_merge_and_unload_set(self) -> bool:
        """Check if the "adapter" configuration section exists and, if affirmative, that it contains the
        "postprocessor" subsection and the "merge_adapter_into_base_model" and "progressbar" directives.

        # Return

            :return (bool): whether merge_and_unload should be done.
        """
        return (
            self.config_obj.adapter is not None
            and self.config_obj.adapter.postprocessor is not None
            and self.config_obj.adapter.postprocessor.merge_adapter_into_base_model
        )

    def merge_and_unload(self, progressbar: bool = False) -> None:
        """This method merges the LoRa layers into the base model.  This is needed if someone wants to use the base
        model as a standalone model.  The implementation calls merge_and_unload() of the underlying LoraModel class
        (in peft).

        Args:
            progressbar (bool): whether to show a progressbar indicating the unload and merge process
        """
        from peft import LoraModel

        if isinstance(self.model.base_model, LoraModel):
            self.model.base_model.merge_and_unload(progressbar=progressbar)
        else:
            raise ValueError("This operation requires an LLM model trained with a LoRA adapter.")

    def _unpack_inputs(
        self,
        inputs: Union[
            Dict[str, torch.Tensor], Dict[str, np.ndarray], Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]
        ],
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """Converts input tensors to input ids."""
        if isinstance(inputs, tuple):
            inputs, targets = inputs
            # Convert targets to tensors.
            for target_feature_name, target_value in targets.items():
                if not isinstance(target_value, torch.Tensor):
                    targets[target_feature_name] = torch.from_numpy(target_value)
                else:
                    targets[target_feature_name] = target_value
        else:
            targets = None

        assert list(inputs.keys()) == self.input_features.keys()

        input_ids = self.get_input_ids(inputs)
        target_ids = self.get_target_ids(targets) if targets else None

        return input_ids, target_ids

    def get_input_ids(
        self,
        inputs: Union[
            Dict[str, torch.Tensor], Dict[str, np.ndarray], Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]
        ],
    ) -> torch.Tensor:
        """Returns the input ids for the text feature input."""
        return inputs[self.config_obj.input_features[0].name].type(torch.int32)

    def get_target_ids(self, outputs: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Returns the output ids for the text feature output."""
        return outputs[self.config_obj.output_features[0].name].type(torch.int32)

    def update_metrics(self, targets, predictions):
        """Updates the model's metrics given targets and predictions for zero-shot/few-shot."""
        for of_name, of_obj in self.output_features.items():
            if isinstance(of_obj, TextOutputFeature):
                # Align the target length with the predictions length to enable text metric evaluation.
                _targets, _predictions = realign_target_and_prediction_tensors_for_inference(
                    targets, predictions, of_name, self.tokenizer
                )
                of_obj.update_metrics(_targets[of_name], _predictions[of_name], self.tokenizer)
            else:
                of_obj.update_metrics(targets[of_name], predictions[of_name])

        # HACK (Tim): get the device of the targets to transfer self.eval_loss_metric to the same device
        target_device = list(targets.values())[0].device

        # HACK (Tim): get the device of the targets to transfer self.eval_loss_metric to the same device
        target_device = list(targets.values())[0].device

        eval_loss, additional_losses = self.eval_loss(targets, predictions)
        self.eval_loss_metric = self.eval_loss_metric.to(target_device)
        self.eval_loss_metric.update(eval_loss)
        self.eval_additional_losses_metrics.update(additional_losses)

    def update_metrics_finetune_llm(self, targets, predictions):
        """Updates the model's metrics given targets and predictions for fine-tuning."""
        _targets, _predictions = targets, predictions
        for of_name, of_obj in self.output_features.items():
            if isinstance(of_obj, TextOutputFeature):
                # Update the target tensor to enable text metric evaluation. This pads the target tensor with -100s
                # to match the prediction length and depends on how much of the target tensor was included in the
                # forward pass.
                _targets = self._update_target_tensor_for_finetuning(_targets, _predictions, of_name)
                if isinstance(of_obj, TextOutputFeature):
                    of_obj.update_metrics(_targets[of_name], _predictions[of_name], self.tokenizer)
                else:
                    of_obj.update_metrics(_targets[of_name], _predictions[of_name])
                continue

            of_obj.update_metrics(_targets[of_name], _predictions[of_name])

        eval_loss, additional_losses = self.eval_loss(_targets, _predictions)
        self.eval_loss_metric.update(eval_loss)
        self.eval_additional_losses_metrics.update(additional_losses)

    def train_loss(
        self,
        targets,
        predictions,
        regularization_type: Optional[str] = None,
        regularization_lambda: Optional[float] = None,
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Computes the training loss for the model.

        Args:
            targets: A dictionary of target names to target tensors.
            predictions: A dictionary of output names to output tensors.
            regularization_type: One of 'l1', 'l2', 'l1_l2', or None.
            regularization_lambda: The regularization lambda.

        Returns:
            A tuple of the loss tensor and a dictionary of loss for every
            output feature.
        """
        train_loss = 0
        of_train_losses = {}
        for of_name, of_obj in self.output_features.items():
            _targets, _predictions = targets, predictions
            if isinstance(of_obj, TextOutputFeature):
                _predictions = {of_name: _predictions}

                # Update the target tensor to enable text metric evaluation. This pads the target tensor with -100s
                # to match the prediction length and depends on how much of the target tensor was included in the
                # forward pass.
                _targets = self._update_target_tensor_for_finetuning(_targets, _predictions, of_name)

            # TODO(Arnav): Seems like doing this again and going between these format types in unnecessary, but
            # refactor so that we don't have to do this at a later point.
            predictions = {}
            for key, _ in _predictions[of_name].items():
                set_output_feature_tensor(predictions, of_name, key, _predictions[of_name][key])
            _predictions = predictions

            of_train_loss = of_obj.train_loss(_targets[of_name], _predictions, of_name)
            train_loss += of_obj.loss.weight * of_train_loss
            of_train_losses[of_name] = of_train_loss

        additional_losses = self.losses()
        if additional_losses:
            train_loss += torch.sum(torch.stack(additional_losses))  # other losses

        # Add regularization loss
        if regularization_type is not None and regularization_lambda != 0:
            train_loss += reg_loss(self, regularization_type, l1=regularization_lambda, l2=regularization_lambda)

        return train_loss, of_train_losses

    def eval_loss(self, targets, predictions):
        """Computes all evaluation losses for the model given targets and predictions.

        Args:
            targets: A dictionary of target names to target tensors.
            predictions: A dictionary of output names to output tensors.

        Returns:
            A tuple of loss values for eval losses and additional losses.
        """
        eval_loss = 0
        for of_name, of_obj in self.output_features.items():
            if isinstance(of_obj, TextOutputFeature):
                # Align the target length with the predictions length to enable text metric evaluation.
                _targets, _predictions = realign_target_and_prediction_tensors_for_inference(
                    targets, predictions, of_name, self.tokenizer
                )
                of_eval_loss = of_obj.eval_loss(_targets[of_name], _predictions[of_name])
            else:
                # HACK(geoffrey): we need a non-empty loss, so we just fill it with zeros
                of_eval_loss = torch.tensor(0.0).to(predictions[of_name][LOGITS].device)

            eval_loss += of_obj.loss.weight * of_eval_loss

        additional_loss = 0
        additional_losses = self.losses()
        if additional_losses:
            additional_loss = torch.sum(torch.stack(additional_losses))  # other losses

        return eval_loss, additional_loss

    def outputs_to_predictions(self, outputs: Dict[str, torch.Tensor]) -> Dict[str, Dict[str, torch.Tensor]]:
        """Returns the model's predictions for each output feature."""
        predictions = {}
        for of_name in self.output_features:
            # TODO(travis): this will need to change when we support multiple output features
            predictions[of_name] = outputs
        return predictions

    def save(self, save_path):
        """Saves the model to the given path."""
        # TODO(travis): use the implementation of trainer itself to decide whether to save the model, to
        # avoid this hack
        if self.config_obj.trainer.type != "none":
            weights_save_path = os.path.join(save_path, MODEL_WEIGHTS_FILE_NAME)
            self.model.save_pretrained(weights_save_path)
        else:
            logger.info("Skipped saving LLM without weight adjustments.")

    def save_base_model(self, save_path):
        """Saves the base LLM model to the given path."""
        # TODO: see the "TODO" statement from "LLM.save()" in this module.
        if self.config_obj.trainer.type != "none":
            weights_save_path = os.path.join(save_path, MODEL_WEIGHTS_FILE_NAME)
            self.model.base_model.save_pretrained(weights_save_path)
            """While this class initializes the tokenizer (from the base_model) automatically, and hence does not
            need to be saved if inference is to be done using LudwigModel.predict(), the rationale for saving the
            tokenizer to HuggingFace Hub is to provide access to models fine-tuned and persisted to HuggingFace Hub
            using Ludwig at a later time, with the ability to perform inference, independently of Ludwig itself."""
            self.tokenizer.save_pretrained(weights_save_path)
        else:
            logger.info("Skipped saving LLM without weight adjustments.")

    def load(self, save_path):
        """Loads the model from the given path."""
        weights_save_path = os.path.join(save_path, MODEL_WEIGHTS_FILE_NAME)
        if self.config_obj.adapter:
            from peft import PeftModel  # noqa

            if isinstance(self.model, PeftModel):
                # Unwrap and reload PeftModel
                self.model = self.model.base_model

            self.model = PeftModel.from_pretrained(self.model, weights_save_path)
        elif self.config_obj.trainer.type != "none":
            self.model = load_pretrained_from_config(
                self.config_obj, model_config=self.model_config, weights_save_path=weights_save_path
            )
        else:
            logger.info("Skipped loading LLM without weight adjustments.")

    def get_args(self):
        """Returns init arguments for constructing this model."""
        return (
            self.config_obj.input_features.to_list(),
            self.config_obj.output_features.to_list(),
            self._random_seed,
        )

    def _update_target_tensor_for_finetuning(
        self, targets: Dict[str, torch.Tensor], predictions: Dict[str, torch.Tensor], of_name: str
    ) -> Dict[str, torch.Tensor]:
        """Update target tensor for fine-tuning.

        This method removes left padding from target tensors, adds a pad token to the end of the target tensors,
        and pads the target tensors with -100 to ensure equal length for loss computation. It then realigns the
        target tensors with the prediction tensors.

        Args:
            targets (Dict[str, torch.Tensor]): A dictionary containing the target tensors.
            predictions (Dict[str, torch.Tensor]): A dictionary containing the predicted tensors.
            of_name (str): The name of the target tensor.

        Returns:
            Dict[str, torch.Tensor]: A dictionary containing the updated target tensors aligned with predictions.
        """
        # Remove left padding from target tensors since we also do this for the model's forward pass when we
        # concatenate the input_ids with the target_ids. We also need to add the pad token to the end of the
        # target tensors.
        targets_without_padding = []
        lengths = []

        pad_token_tensor = torch.tensor([self.tokenizer.pad_token_id])
        for target in targets[of_name]:
            target = remove_left_padding(target, self.tokenizer)[0]
            target = torch.cat([target, pad_token_tensor.to(device=target.device)], dim=-1).unsqueeze(0)
            targets_without_padding.append(target)
            lengths.append(target.shape[1])

        # We need all target tensors to have the same length for the loss computation. We pad the target
        # tensors with -100 since we want to negate all tokens that are not target_ids during the softmax
        # cross entropy loss computation. This ensures that the loss is computed only for the target tokens.
        max_length = max(lengths)
        for i, target in enumerate(targets_without_padding):
            targets_without_padding[i] = add_left_padding(
                targets_without_padding[i][0],
                max_length,
                IGNORE_INDEX_TOKEN_ID,
            )

        targets[of_name] = torch.stack(targets_without_padding, dim=0).to(
            dtype=targets[of_name].dtype,
            device=targets[of_name].device,
        )

        # Re-align target tensors without padding to have equal length before realigning with the prediction
        # tensors. Padding left with -100 to match the length of the target tensor masks the input ids during
        # softmax cross entropy loss computation. This ensures that the loss is computed only for the target
        # token IDs. Examples:
        # BERTLMHead: https://github.com/huggingface/transformers/blob/v4.29.1/src/transformers/models/bert/modeling_bert.py#L1216-L1219 # noqa
        # GPTNeoForCausalLM: https://github.com/huggingface/transformers/blob/v4.29.1/src/transformers/models/gpt_neo/modeling_gpt_neo.py#L736 # noqa
        _targets = pad_target_tensor_for_fine_tuning(targets, predictions, self.model_inputs, of_name)

        return _targets

    @staticmethod
    def get_augmentation_pipelines() -> AugmentationPipelines:
        """Returns the augmentation pipeline for this model."""
        return AugmentationPipelines({})



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Type

import torch
from torch import nn, Tensor
from torch.nn import HuberLoss as _HuberLoss
from torch.nn import L1Loss
from torch.nn import MSELoss as _MSELoss
from torchmetrics.functional import mean_absolute_percentage_error

import ludwig.utils.loss_utils as utils
from ludwig.constants import LOGITS
from ludwig.modules.loss_implementations.corn import corn_loss
from ludwig.schema.features.loss.loss import (
    BaseLossConfig,
    BWCEWLossConfig,
    CORNLossConfig,
    HuberLossConfig,
    MAELossConfig,
    MAPELossConfig,
    MSELossConfig,
    NextTokenSoftmaxCrossEntropyLossConfig,
    RMSELossConfig,
    RMSPELossConfig,
    SequenceSoftmaxCrossEntropyLossConfig,
    SigmoidCrossEntropyLossConfig,
    SoftmaxCrossEntropyLossConfig,
)
from ludwig.utils import strings_utils
from ludwig.utils.registry import Registry

# used for Laplace smoothing for candidate samplers
EPSILON = 1.0e-10

loss_impl_registry = Registry[Type[nn.Module]]()


def register_loss(config_cls: Type[BaseLossConfig]):
    def wrap(cls: Type[nn.Module]):
        loss_impl_registry[config_cls] = cls
        return cls

    return wrap


def create_loss(config: BaseLossConfig) -> nn.Module:
    return loss_impl_registry[type(config)](config)


class LogitsInputsMixin:
    @classmethod
    def get_loss_inputs(cls):
        """Maps loss to the desired predicted input type."""
        return LOGITS


@register_loss(MSELossConfig)
class MSELoss(_MSELoss, LogitsInputsMixin):
    """Mean squared error."""

    def __init__(self, config: MSELossConfig):
        super().__init__()


@register_loss(MAELossConfig)
class MAELoss(L1Loss, LogitsInputsMixin):
    """Mean absolute error."""

    def __init__(self, config: MAELossConfig):
        super().__init__()


@register_loss(MAPELossConfig)
class MAPELoss(nn.Module, LogitsInputsMixin):
    """Mean absolute error."""

    def __init__(self, config: MAPELossConfig):
        super().__init__()

    def forward(self, preds: Tensor, target: Tensor) -> Tensor:
        return mean_absolute_percentage_error(preds, target)


@register_loss(RMSELossConfig)
class RMSELoss(nn.Module, LogitsInputsMixin):
    """Root mean square error."""

    def __init__(self, config: RMSELossConfig):
        super().__init__()
        self.mse = nn.MSELoss()

    def forward(self, preds: Tensor, target: Tensor) -> Tensor:
        return torch.sqrt(self.mse(preds, target))


@register_loss(RMSPELossConfig)
class RMSPELoss(nn.Module, LogitsInputsMixin):
    """Root mean square percentage error."""

    def __init__(self, config: RMSPELossConfig):
        super().__init__()

    def forward(self, preds: Tensor, target: Tensor) -> Tensor:
        loss = utils.rmspe_loss(target, preds)
        return loss


@register_loss(BWCEWLossConfig)
class BWCEWLoss(nn.Module, LogitsInputsMixin):
    """Binary weighted cross entropy loss."""

    def __init__(self, config: BWCEWLossConfig):
        super().__init__()
        if config.positive_class_weight:
            self.loss_fn = nn.BCEWithLogitsLoss(pos_weight=torch.Tensor([config.positive_class_weight]))
        else:
            self.loss_fn = nn.BCEWithLogitsLoss(pos_weight=config.positive_class_weight)
        self.robust_lambda = config.robust_lambda
        self.confidence_penalty = config.confidence_penalty

    def forward(self, preds: torch.Tensor, target: torch.Tensor):
        train_loss = self.loss_fn(preds, target.float())
        # robust lambda
        if self.robust_lambda > 0:
            train_loss = (1 - self.robust_lambda) * train_loss + self.robust_lambda / 2

        train_mean_loss = torch.mean(train_loss)

        # confidence penalty
        if self.confidence_penalty > 0:
            probabilities = torch.sigmoid(preds)
            mean_penalty = utils.mean_confidence_penalty(probabilities, 2)
            train_mean_loss += self.confidence_penalty * mean_penalty

        return train_mean_loss


@register_loss(SoftmaxCrossEntropyLossConfig)
class SoftmaxCrossEntropyLoss(nn.Module, LogitsInputsMixin):
    def __init__(self, config: SoftmaxCrossEntropyLossConfig):
        """
        Params:
            class_weights: List or 1D tensor of length equal to number of classes.
        """
        super().__init__()
        if config.class_weights:
            self.loss_fn = nn.CrossEntropyLoss(weight=torch.Tensor(config.class_weights))
        else:
            self.loss_fn = nn.CrossEntropyLoss()

    def forward(self, preds: Tensor, target: Tensor) -> Tensor:
        """
        Params:
            preds: Tensor of shape [batch x num_classes]
            target: Tensor of shape [batch], where each element is integral
                between 0 and num_classes.
        """
        if len(target.shape) == 1:
            # Assumes we are providing the target as a single class, rather than a distribution
            target = target.long()
        return self.loss_fn(preds, target)


@register_loss(SequenceSoftmaxCrossEntropyLossConfig)
class SequenceSoftmaxCrossEntropyLoss(nn.Module, LogitsInputsMixin):
    def __init__(self, config: SequenceSoftmaxCrossEntropyLossConfig):
        """
        Params:
            class_weights: List or 1D tensor of length equal to number of classes.
        """
        super().__init__()
        if config.class_weights:
            self.loss_fn = nn.CrossEntropyLoss(
                weight=torch.Tensor(config.class_weights), ignore_index=strings_utils.SpecialSymbol.PADDING.value
            )
        else:
            self.loss_fn = nn.CrossEntropyLoss(ignore_index=strings_utils.SpecialSymbol.PADDING.value)

    def forward(self, preds: Tensor, target: Tensor) -> Tensor:
        """
        Params:
            preds: Tensor of shape [batch x sequence_length x vocab_size]
            target: Tensor of shape [batch x sequence_length], where each element is integral between 0 and vocab_size.
        """
        target = target.long()
        return self.loss_fn(preds[1:].view(-1, preds.size(-1)), target[1:].view(-1))


@register_loss(NextTokenSoftmaxCrossEntropyLossConfig)
class NextTokenSoftmaxCrossEntropyLoss(nn.Module, LogitsInputsMixin):
    def __init__(self, config: NextTokenSoftmaxCrossEntropyLossConfig):
        super().__init__()
        self.loss_fn = nn.CrossEntropyLoss()

    def forward(self, preds: Tensor, target: Tensor) -> Tensor:
        """
        Params:
            preds: Tensor of shape [batch x sequence_length x vocab_size]
            target: Tensor of shape [batch x sequence_length], where each element is integral between 0 and vocab_size.

        Reference implementation:
        https://github.com/huggingface/transformers/blob/v4.29.1/src/transformers/models/bert/modeling_bert.py#LL1253C1-L1260C1 # noqa
        """
        target = target.long()
        _, _, vocab_size = preds.shape
        # logits for all tensors except n+1 since each logit tensor at position i represents the log probabilities for
        # the next token i+1 if we were to do argmax on the logits ensor at position i.
        shifted_predictions = preds[:, :-1, :]
        # Shift by 1 since the logits at position 0 in predictions represent the log likelihood of target token 1
        shifted_targets = target[:, 1:]
        return self.loss_fn(shifted_predictions.reshape(-1, vocab_size), shifted_targets.reshape(-1))


@register_loss(SigmoidCrossEntropyLossConfig)
class SigmoidCrossEntropyLoss(nn.Module, LogitsInputsMixin):
    def __init__(self, config: SigmoidCrossEntropyLossConfig):
        """
        Params:
            class_weights: List or 1D tensor of length equal to number of classes.
        """
        super().__init__()
        if config.class_weights:
            self.loss_fn = nn.BCEWithLogitsLoss(pos_weight=torch.Tensor(config.class_weights))
        else:
            self.loss_fn = nn.BCEWithLogitsLoss()

    def forward(self, preds: Tensor, target: Tensor) -> Tensor:
        if preds.ndim != 2:
            raise RuntimeError("SigmoidCrossEntropyLoss currently only supported for 2D tensors.")

        return self.loss_fn(preds.type(torch.float32), target.type(torch.float32))


@register_loss(HuberLossConfig)
class HuberLoss(_HuberLoss, LogitsInputsMixin):
    """Huber loss."""

    def __init__(self, config: HuberLossConfig):
        super().__init__(delta=config.delta)


@register_loss(CORNLossConfig)
class CORNLoss(nn.Module, LogitsInputsMixin):
    """CORN loss."""

    def __init__(self, config: CORNLossConfig):
        super().__init__()

    def forward(self, preds: Tensor, target: Tensor) -> Tensor:
        num_classes = preds.shape[1]
        return corn_loss(preds, target, num_classes=num_classes)



import argparse
import logging
import sys
from typing import List, Optional, Union

import pandas as pd

from ludwig.api import LudwigModel
from ludwig.backend import ALL_BACKENDS, Backend, initialize_backend
from ludwig.callbacks import Callback
from ludwig.contrib import add_contrib_callback_args
from ludwig.globals import LUDWIG_VERSION
from ludwig.utils.print_utils import get_logging_level_registry, print_ludwig

logger = logging.getLogger(__name__)


def forecast_cli(
    model_path: str,
    dataset: Union[str, dict, pd.DataFrame] = None,
    data_format: Optional[str] = None,
    horizon: int = 1,
    output_directory: Optional[str] = None,
    output_format: str = "parquet",
    callbacks: List[Callback] = None,
    backend: Union[Backend, str] = None,
    logging_level: int = logging.INFO,
    **kwargs,
) -> None:
    """Loads pre-trained model to forecast on the provided dataset.

    # Inputs

    :param model_path: (str) filepath to pre-trained model.
    :param dataset: (Union[str, dict, pandas.DataFrame], default: `None`)
        source containing the entire dataset to be used in the prediction.
    :param data_format: (str, default: `None`) format to interpret data
        sources. Will be inferred automatically if not specified.
    :param horizon: How many samples into the future to forecast.
    :param output_directory: (str, default: `'results'`) the directory that
        will contain the forecasted values.
    :param output_format: (str) format of the output dataset.
    :param callbacks: (list, default: `None`) a list of
        `ludwig.callbacks.Callback` objects that provide hooks into the
        Ludwig pipeline.
    :param backend: (Union[Backend, str]) `Backend` or string name
        of backend to use to execute preprocessing / training steps.
    :param logging_level: (int) Log level that will be sent to stderr.

    # Returns

    :return: ('None')
    """
    model = LudwigModel.load(
        model_path,
        logging_level=logging_level,
        backend=backend,
        callbacks=callbacks,
    )
    model.forecast(
        dataset=dataset,
        data_format=data_format,
        horizon=horizon,
        output_directory=output_directory,
        output_format=output_format,
    )


def cli(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script loads a pretrained model and uses it to forecast",
        prog="ludwig forecast",
        usage="%(prog)s [options]",
    )

    parser.add_argument(
        "-n", "--horizon", help="horizon, or number of steps in the future to forecast", type=int, default=1
    )

    # ---------------
    # Data parameters
    # ---------------
    parser.add_argument("--dataset", help="input data file path", required=True)
    parser.add_argument(
        "--data_format",
        help="format of the input data",
        default="auto",
        choices=[
            "auto",
            "csv",
            "excel",
            "feather",
            "fwf",
            "hdf5",
            "html",
            "tables",
            "json",
            "jsonl",
            "parquet",
            "pickle",
            "sas",
            "spss",
            "stata",
            "tsv",
        ],
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument("-m", "--model_path", help="model to load", required=True)

    # -------------------------
    # Output results parameters
    # -------------------------
    parser.add_argument(
        "-od", "--output_directory", type=str, default="results", help="directory that contains the results"
    )

    parser.add_argument(
        "-of",
        "--output_format",
        help="format to write the output dataset",
        default="parquet",
        choices=[
            "csv",
            "parquet",
        ],
    )

    parser.add_argument(
        "-b",
        "--backend",
        help="specifies backend to use for parallel / distributed execution, "
        "defaults to local execution or Horovod if called using horovodrun",
        choices=ALL_BACKENDS,
    )

    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="the level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("forecast", *sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.forecast")

    args.backend = initialize_backend(args.backend)
    if args.backend.is_coordinator():
        print_ludwig("Forecast", LUDWIG_VERSION)
        logger.info(f"Dataset path: {args.dataset}")
        logger.info(f"Model path: {args.model_path}")
        logger.info("")

    forecast_cli(**vars(args))


if __name__ == "__main__":
    cli(sys.argv[1:])



import os
from typing import Type

import pytest
from transformers import AlbertModel, BertModel, BertTokenizer

from ludwig.encoders.text_encoders import ALBERTEncoder, BERTEncoder
from ludwig.utils.hf_utils import load_pretrained_hf_model_from_hub, load_pretrained_hf_model_with_hub_fallback


@pytest.mark.parametrize(
    ("model", "name"),
    [
        (AlbertModel, ALBERTEncoder.DEFAULT_MODEL_NAME),
        (BertTokenizer, "bert-base-uncased"),
    ],
)
def test_load_pretrained_hf_model_from_hub(model: Type, name: str, tmpdir: os.PathLike):
    """Ensure that the HF models used in ludwig download correctly."""
    cache_dir = os.path.join(tmpdir, name.replace(os.path.sep, "_") if name else str(model.__name__))
    os.makedirs(cache_dir, exist_ok=True)
    loaded_model = load_pretrained_hf_model_from_hub(model, name, cache_dir=cache_dir, force_download=True)
    assert isinstance(loaded_model, model)
    assert os.listdir(cache_dir)


def test_load_pretrained_hf_model_with_hub_fallback(tmpdir):
    """Ensure that the HF models used in ludwig download correctly with S3 or hub fallback."""
    # Don't set env var.
    _, used_fallback = load_pretrained_hf_model_with_hub_fallback(AlbertModel, ALBERTEncoder.DEFAULT_MODEL_NAME)
    assert used_fallback

    # Download the model, load it from tmpdir, and set env var.
    load_pretrained_hf_model_from_hub(AlbertModel, "albert-base-v2").save_pretrained(
        os.path.join(tmpdir, "albert-base-v2")
    )
    os.environ["LUDWIG_PRETRAINED_MODELS_DIR"] = f"file://{tmpdir}"  # Needs to be an absolute path.
    _, used_fallback = load_pretrained_hf_model_with_hub_fallback(AlbertModel, ALBERTEncoder.DEFAULT_MODEL_NAME)
    assert not used_fallback

    # Fallback is used for a model that doesn't exist in models directory.
    _, used_fallback = load_pretrained_hf_model_with_hub_fallback(BertModel, BERTEncoder.DEFAULT_MODEL_NAME)
    assert used_fallback

    # Clean up.
    del os.environ["LUDWIG_PRETRAINED_MODELS_DIR"]



import logging
from typing import Dict, List

from tabulate import tabulate

from ludwig.constants import COMBINED, LOSS
from ludwig.utils.metric_utils import TrainerMetric

logger = logging.getLogger(__name__)


def get_metric_value_or_empty(metrics_log: Dict[str, List[TrainerMetric]], metric_name: str):
    """Returns the metric value if it exists or empty."""
    if metric_name not in metrics_log:
        return ""
    return metrics_log[metric_name][-1][-1]


def print_table_for_single_output_feature(
    train_metrics_log: Dict[str, List[TrainerMetric]],
    validation_metrics_log: Dict[str, List[TrainerMetric]],
    test_metrics_log: Dict[str, List[TrainerMetric]],
    combined_loss_for_each_split: List[float],
) -> None:
    """Prints the metrics table for a single output feature.

    Args:
        train_metrics_log: Dict from metric name to list of TrainerMetric.
        validation_metrics_log: Dict from metric name to list of TrainerMetric.
        test_metrics_log: Dict from metric name to list of TrainerMetric.
    """
    # Get the superset of metric names across all splits.
    all_metric_names = set()
    all_metric_names.update(train_metrics_log.keys())
    all_metric_names.update(validation_metrics_log.keys())
    all_metric_names.update(test_metrics_log.keys())
    all_metric_names = sorted(list(all_metric_names))

    # Assemble the printed table.
    # Each item in the printed_table corresponds to a row in the printed table.
    printed_table = [["train", "validation", "test"]]
    for metric_name in all_metric_names:
        metrics_for_each_split = [
            get_metric_value_or_empty(train_metrics_log, metric_name),
            get_metric_value_or_empty(validation_metrics_log, metric_name),
            get_metric_value_or_empty(test_metrics_log, metric_name),
        ]
        printed_table.append([metric_name] + metrics_for_each_split)

    # Add combined loss.
    printed_table.append(["combined_loss"] + combined_loss_for_each_split)

    logger.info(tabulate(printed_table, headers="firstrow", tablefmt="fancy_grid", floatfmt=".4f"))


def print_metrics_table(
    output_features: Dict[str, "OutputFeature"],  # noqa
    train_metrics_log: Dict[str, Dict[str, List[TrainerMetric]]],
    validation_metrics_log: Dict[str, Dict[str, List[TrainerMetric]]],
    test_metrics_log: Dict[str, Dict[str, List[TrainerMetric]]],
):
    """Prints a table of metrics table for each output feature, for each split.

    Example:
    ╒═══════════════╤═════════╤══════════════╤════════╕
    │               │   train │   validation │   test │
    ╞═══════════════╪═════════╪══════════════╪════════╡
    │ accuracy      │  0.8157 │       0.6966 │ 0.8090 │
    ├───────────────┼─────────┼──────────────┼────────┤
    │ loss          │  0.4619 │       0.5039 │ 0.4488 │
    ├───────────────┼─────────┼──────────────┼────────┤
    │ precision     │  0.8274 │       0.6250 │ 0.7818 │
    ├───────────────┼─────────┼──────────────┼────────┤
    │ recall        │  0.6680 │       0.4545 │ 0.6615 │
    ├───────────────┼─────────┼──────────────┼────────┤
    │ roc_auc       │  0.8471 │       0.7706 │ 0.8592 │
    ├───────────────┼─────────┼──────────────┼────────┤
    │ specificity   │  0.9105 │       0.8393 │ 0.8938 │
    ├───────────────┼─────────┼──────────────┼────────┤
    │ combined_loss │  0.4619 │       0.5039 │ 0.4488 │
    ╘═══════════════╧═════════╧══════════════╧════════╛
    """
    # Obtain the combined loss, which is the same across all output features.
    combined_loss_for_each_split = [
        get_metric_value_or_empty(train_metrics_log[COMBINED], LOSS),
        get_metric_value_or_empty(validation_metrics_log[COMBINED], LOSS),
        get_metric_value_or_empty(test_metrics_log[COMBINED], LOSS),
    ]

    for output_feature_name in sorted(output_features.keys()):
        if output_feature_name == COMBINED:
            # Skip the combined output feature. The combined loss will be added to each output feature's table.
            continue
        print_table_for_single_output_feature(
            train_metrics_log[output_feature_name],
            validation_metrics_log[output_feature_name],
            test_metrics_log[output_feature_name],
            combined_loss_for_each_split,
        )



from typing import TypedDict

import torch


class EncoderOutputDict(TypedDict, total=False):
    encoder_output: torch.Tensor
    encoder_output_state: torch.Tensor  # only used by sequence and h3 encoders
    attentions: torch.Tensor  # only used by the vit legacy encoder



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandas as pd

from ludwig.datasets.loaders.dataset_loader import DatasetLoader


class ConsumerComplaintsLoader(DatasetLoader):
    """The Consumer Complaints dataset."""

    def load_file_to_dataframe(self, file_path: str) -> pd.DataFrame:
        """Loads a file into a dataframe."""

        consumer_complaints_df = pd.read_csv(file_path)
        consumer_complaints_df = preprocess_df(consumer_complaints_df)

        return consumer_complaints_df


def preprocess_df(df):
    """Preprocesses the dataframe.

        - Remove all rows with missing values in the following columns:
            - Consumer complaint narrative
            - Issue
            - Product

    Args:
        df (pd.DataFrame): The dataframe to preprocess.

    Returns:
        pd.DataFrame: The preprocessed dataframe.
    """
    return df.dropna(subset=["Consumer complaint narrative", "Issue", "Product"])



"""Uses heuristics to build ludwig configuration file:

(1) infer types based on dataset
(2) populate with
    - default combiner parameters,
    - preprocessing parameters,
    - combiner specific default training parameters,
    - combiner specific hyperopt space
    - feature parameters
(3) add machineresources
    (base implementation -- # CPU, # GPU)
"""

import logging
import os
from dataclasses import dataclass
from typing import Any, Dict, List, Set, Union

import dask.dataframe as dd
import numpy as np
import pandas as pd
import yaml
from dataclasses_json import dataclass_json, LetterCase
from tqdm import tqdm

from ludwig.api_annotations import DeveloperAPI
from ludwig.backend import Backend
from ludwig.constants import (
    COLUMN,
    COMBINER,
    ENCODER,
    EXECUTOR,
    HYPEROPT,
    INPUT_FEATURES,
    PREPROCESSING,
    SCHEDULER,
    SEARCH_ALG,
    SPLIT,
    TEXT,
    TYPE,
)
from ludwig.types import ModelConfigDict
from ludwig.utils.automl.data_source import DataSource, wrap_data_source
from ludwig.utils.automl.field_info import FieldConfig, FieldInfo, FieldMetadata
from ludwig.utils.automl.type_inference import infer_type, should_exclude
from ludwig.utils.data_utils import load_yaml
from ludwig.utils.misc_utils import merge_dict
from ludwig.utils.system_utils import Resources

logger = logging.getLogger(__name__)

PATH_HERE = os.path.abspath(os.path.dirname(__file__))
CONFIG_DIR = os.path.join(PATH_HERE, "defaults")

BASE_AUTOML_CONFIG = os.path.join(CONFIG_DIR, "base_automl_config.yaml")
REFERENCE_CONFIGS = os.path.join(CONFIG_DIR, "reference_configs.yaml")

combiner_defaults = {
    "concat": os.path.join(CONFIG_DIR, "combiner/concat_config.yaml"),
    "tabnet": os.path.join(CONFIG_DIR, "combiner/tabnet_config.yaml"),
    "transformer": os.path.join(CONFIG_DIR, "combiner/transformer_config.yaml"),
}

encoder_defaults = {"text": {"bert": os.path.join(CONFIG_DIR, "text/bert_config.yaml")}}

# Cap for number of distinct values to return.
MAX_DISTINCT_VALUES_TO_RETURN = 10


@DeveloperAPI
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DatasetInfo:
    fields: List[FieldInfo]
    row_count: int
    size_bytes: int = -1


def allocate_experiment_resources(resources: Resources) -> dict:
    """Allocates ray trial resources based on available resources.

    # Inputs
    :param resources (dict) specifies all available GPUs, CPUs and associated
        metadata of the machines (i.e. memory)

    # Return
    :return: (dict) gpu and cpu resources per trial
    """
    # TODO (ASN):
    # (1) expand logic to support multiple GPUs per trial (multi-gpu training)
    # (2) add support for kubernetes namespace (if applicable)
    # (3) add support for smarter allocation based on size of GPU memory
    experiment_resources = {"cpu_resources_per_trial": 1}
    gpu_count, cpu_count = resources.gpus, resources.cpus
    if gpu_count > 0:
        experiment_resources.update({"gpu_resources_per_trial": 1})
        if cpu_count > 1:
            cpus_per_trial = max(int(cpu_count / gpu_count), 1)
            experiment_resources["cpu_resources_per_trial"] = cpus_per_trial

    return experiment_resources


def get_resource_aware_hyperopt_config(
    experiment_resources: Dict[str, Any], time_limit_s: Union[int, float], random_seed: int
) -> Dict[str, Any]:
    """Returns a Ludwig config with the hyperopt section populated with appropriate parameters.

    Hyperopt parameters are intended to be appropriate for the given resources and time limit.
    """
    executor = experiment_resources
    executor.update({"time_budget_s": time_limit_s})
    if time_limit_s is not None:
        executor.update({SCHEDULER: {"max_t": time_limit_s}})

    return {
        HYPEROPT: {
            SEARCH_ALG: {"random_state_seed": random_seed},
            EXECUTOR: executor,
        },
    }


def _get_stratify_split_config(field_meta: FieldMetadata) -> dict:
    return {
        PREPROCESSING: {
            SPLIT: {
                TYPE: "stratify",
                COLUMN: field_meta.name,
            }
        }
    }


def get_default_automl_hyperopt() -> Dict[str, Any]:
    """Returns general, default settings for hyperopt.

    For example:
    - We set a random_state_seed for sample sequence repeatability
    - We use an increased reduction_factor to get more pruning/exploration.

    TODO: If settings seem reasonable, consider building this into the hyperopt schema, directly.
    """
    return yaml.safe_load(
        """
  search_alg:
    type: hyperopt
  executor:
    type: ray
    num_samples: 10
    time_budget_s: 3600
    scheduler:
      type: async_hyperband
      time_attr: time_total_s
      max_t: 3600
      grace_period: 72
      reduction_factor: 5
"""
    )


def create_default_config(
    features_config: ModelConfigDict,
    dataset_info: DatasetInfo,
    target_name: Union[str, List[str]],
    time_limit_s: Union[int, float],
    random_seed: int,
    imbalance_threshold: float = 0.9,
    backend: Backend = None,
) -> dict:
    """Returns auto_train configs for three available combiner models. Coordinates the following tasks:

    - extracts fields and generates list of FieldInfo objects
    - gets field metadata (i.e avg. words, total non-null entries)
    - builds input_features and output_features section of config
    - for imbalanced datasets, a preprocessing section is added to perform stratified sampling if the imbalance ratio
      is smaller than imbalance_threshold
    - for each combiner, adds default training, hyperopt
    - infers resource constraints and adds gpu and cpu resource allocation per
      trial

    # Inputs
    :param dataset_info: (str) filepath Dataset Info object.
    :param target_name: (str, List[str]) name of target feature
    :param time_limit_s: (int, float) total time allocated to auto_train. acts
                                    as the stopping parameter
    :param random_seed: (int, default: `42`) a random seed that will be used anywhere
                        there is a call to a random number generator, including
                        hyperparameter search sampling, as well as data splitting,
                        parameter initialization and training set shuffling
    :param imbalance_threshold: (float) maximum imbalance ratio (minority / majority) to perform stratified sampling
    :param backend: (Backend) backend to use for training.

    # Return
    :return: (dict) dictionaries contain auto train config files for all available
    combiner types
    """
    base_automl_config = load_yaml(BASE_AUTOML_CONFIG)
    base_automl_config.update(features_config)

    targets = convert_targets(target_name)
    features_metadata = get_field_metadata(dataset_info.fields, dataset_info.row_count, targets)

    # Handle expensive features for CPU
    resources = backend.get_available_resources()
    for ifeature in base_automl_config[INPUT_FEATURES]:
        if resources.gpus == 0:
            if ifeature[TYPE] == TEXT:
                # When no GPUs are available, default to the embed encoder, which is fast enough for CPU
                ifeature[ENCODER] = {"type": "embed"}

    # create set of all feature types appearing in the dataset
    feature_types = [[feat[TYPE] for feat in features] for features in features_config.values()]
    feature_types = set(sum(feature_types, []))

    model_configs = {}

    # update hyperopt config
    experiment_resources = allocate_experiment_resources(resources)
    base_automl_config = merge_dict(
        base_automl_config, get_resource_aware_hyperopt_config(experiment_resources, time_limit_s, random_seed)
    )

    # add preprocessing section if single output feature is imbalanced
    outputs_metadata = [f for f in features_metadata if f.mode == "output"]
    if len(outputs_metadata) == 1:
        of_meta = outputs_metadata[0]
        is_categorical = of_meta.config.type in ["category", "binary"]
        is_imbalanced = of_meta.imbalance_ratio < imbalance_threshold
        if is_categorical and is_imbalanced:
            base_automl_config.update(_get_stratify_split_config(of_meta))

    model_configs["base_config"] = base_automl_config

    # read in all encoder configs
    for feat_type, default_configs in encoder_defaults.items():
        if feat_type in feature_types:
            if feat_type not in model_configs.keys():
                model_configs[feat_type] = {}
            for encoder_name, encoder_config_path in default_configs.items():
                model_configs[feat_type][encoder_name] = load_yaml(encoder_config_path)

    # read in all combiner configs
    model_configs[COMBINER] = {}
    for combiner_type, default_config in combiner_defaults.items():
        combiner_config = load_yaml(default_config)
        model_configs[COMBINER][combiner_type] = combiner_config

    return model_configs


# Read in the score and configuration of a reference model trained by Ludwig for each dataset in a list.
def get_reference_configs() -> dict:
    reference_configs = load_yaml(REFERENCE_CONFIGS)
    return reference_configs


def get_dataset_info(df: Union[pd.DataFrame, dd.core.DataFrame]) -> DatasetInfo:
    """Constructs FieldInfo objects for each feature in dataset. These objects are used for downstream type
    inference.

    # Inputs
    :param df: (Union[pd.DataFrame, dd.core.DataFrame]) Pandas or Dask dataframe.

    # Return
    :return: (DatasetInfo) Structure containing list of FieldInfo objects.
    """
    source = wrap_data_source(df)
    return get_dataset_info_from_source(source)


def is_field_boolean(source: DataSource, field: str) -> bool:
    """Returns a boolean indicating whether the object field should have a bool dtype.

    Columns with object dtype that have 3 distinct values of which one is Nan/None is a bool type column.
    """
    unique_values = source.df[field].unique()
    if len(unique_values) <= 3:
        for entry in unique_values:
            try:
                if np.isnan(entry):
                    continue
            except TypeError:
                # For some field types such as object arrays, np.isnan throws a TypeError
                # In this case, do nothing and proceed to checking if the entry is a bool object
                pass
            if isinstance(entry, bool):
                continue
            return False
        return True
    return False


@DeveloperAPI
def get_dataset_info_from_source(source: DataSource) -> DatasetInfo:
    """Constructs FieldInfo objects for each feature in dataset. These objects are used for downstream type
    inference.

    # Inputs
    :param source: (DataSource) A wrapper around a data source, which may represent a pandas or Dask dataframe.

    # Return
    :return: (DatasetInfo) Structure containing list of FieldInfo objects.
    """
    row_count = len(source)
    fields = []
    for field in tqdm(source.columns, desc="Analyzing fields", total=len(source.columns)):
        logger.info(f"Analyzing field: {field}")
        dtype = source.get_dtype(field)
        num_distinct_values, distinct_values, distinct_values_balance = source.get_distinct_values(
            field, MAX_DISTINCT_VALUES_TO_RETURN
        )
        nonnull_values = source.get_nonnull_values(field)
        image_values = source.get_image_values(field)
        audio_values = source.get_audio_values(field)

        if dtype == "object":
            # Check if it is a nullboolean field. We do this since if you read a csv with
            # pandas that has a column of booleans and some missing values, the column is
            # interpreted as object dtype instead of bool
            if is_field_boolean(source, field):
                dtype = "bool"

        avg_words = None
        if source.is_string_type(dtype):
            try:
                avg_words = source.get_avg_num_tokens(field)
            except AttributeError:
                # Series is not actually a string type despite being an object, e.g., Decimal, Datetime, etc.
                avg_words = None

        fields.append(
            FieldInfo(
                name=field,
                dtype=dtype,
                distinct_values=distinct_values,
                num_distinct_values=num_distinct_values,
                distinct_values_balance=distinct_values_balance,
                nonnull_values=nonnull_values,
                image_values=image_values,
                audio_values=audio_values,
                avg_words=avg_words,
            )
        )
    return DatasetInfo(fields=fields, row_count=row_count, size_bytes=source.size_bytes())


def get_features_config(
    fields: List[FieldInfo],
    row_count: int,
    target_name: Union[str, List[str]] = None,
) -> dict:
    """Constructs FieldInfo objects for each feature in dataset. These objects are used for downstream type
    inference.

    # Inputs
    :param fields: (List[FieldInfo]) FieldInfo objects for all fields in dataset
    :param row_count: (int) total number of entries in original dataset
    :param target_name (str, List[str]) name of target feature

    # Return
    :return: (dict) section of auto_train config for input_features and output_features
    """
    targets = convert_targets(target_name)
    metadata = get_field_metadata(fields, row_count, targets)
    return get_config_from_metadata(metadata, targets)


def convert_targets(target_name: Union[str, List[str]] = None) -> Set[str]:
    targets = target_name
    if isinstance(targets, str):
        targets = [targets]
    if targets is None:
        targets = []
    return set(targets)


def get_config_from_metadata(metadata: List[FieldMetadata], targets: Set[str] = None) -> dict:
    """Builds input/output feature sections of auto-train config using field metadata.

    # Inputs
    :param metadata: (List[FieldMetadata]) field descriptions
    :param targets (Set[str]) names of target features

    # Return
    :return: (dict) section of auto_train config for input_features and output_features
    """
    config = {
        "input_features": [],
        "output_features": [],
    }

    for field_meta in metadata:
        if field_meta.name in targets:
            config["output_features"].append(field_meta.config.to_dict())
        elif not field_meta.excluded and field_meta.mode == "input":
            config["input_features"].append(field_meta.config.to_dict())

    return config


@DeveloperAPI
def get_field_metadata(fields: List[FieldInfo], row_count: int, targets: Set[str] = None) -> List[FieldMetadata]:
    """Computes metadata for each field in dataset.

    # Inputs
    :param fields: (List[FieldInfo]) FieldInfo objects for all fields in dataset
    :param row_count: (int) total number of entries in original dataset
    :param targets (Set[str]) names of target features

    # Return
    :return: (List[FieldMetadata]) list of objects containing metadata for each field
    """

    metadata = []
    column_count = len(fields)
    for idx, field in enumerate(fields):
        missing_value_percent = 1 - float(field.nonnull_values) / row_count
        dtype = infer_type(field, missing_value_percent, row_count)
        metadata.append(
            FieldMetadata(
                name=field.name,
                config=FieldConfig(
                    name=field.name,
                    column=field.name,
                    type=dtype,
                ),
                excluded=should_exclude(idx, field, dtype, column_count, row_count, targets),
                mode=infer_mode(field, targets),
                missing_values=missing_value_percent,
                imbalance_ratio=field.distinct_values_balance,
            )
        )

    return metadata


def infer_mode(field: FieldInfo, targets: Set[str] = None) -> str:
    if field.name in targets:
        return "output"
    if field.name.lower() == "split":
        return "split"
    return "input"



from abc import ABC, abstractmethod
from typing import Optional, Type, TYPE_CHECKING

from ludwig.api_annotations import DeveloperAPI
from ludwig.error import ConfigValidationError
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import LLM_METADATA
from ludwig.schema.metadata.parameter_metadata import convert_metadata_to_json
from ludwig.schema.utils import ludwig_dataclass
from ludwig.utils.registry import Registry

if TYPE_CHECKING:
    from peft import PeftConfig


adapter_registry = Registry()


@DeveloperAPI
def register_adapter(name: str):
    def wrap(config: BaseAdapterConfig):
        adapter_registry[name] = config
        return config

    return wrap


@DeveloperAPI
@ludwig_dataclass
class LoraPostprocessorConfig(schema_utils.BaseMarshmallowConfig):
    """This Dataclass is a schema for the nested postprocessing config under adapter of type "lora"."""

    merge_adapter_into_base_model: bool = schema_utils.Boolean(
        default=False,
        description="""Instructs whether or not the fine-tuned LoRA weights are to be merged into the base LLM model so
that the complete fine-tuned model is available to be used and/or persisted, and then reused upon loading as a single
model (rather than having to load base and fine-tuned models separately).""",
    )
    progressbar: bool = schema_utils.Boolean(
        default=False,
        description="Instructs whether or not to show a progress bar indicating the unload and merge process.",
    )


@DeveloperAPI
class LoraPostprocessorConfigField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(LoraPostprocessorConfig)

    def _jsonschema_type_mapping(self):
        return schema_utils.unload_jsonschema_from_marshmallow_class(LoraPostprocessorConfig, title="LoraPostprocessor")


@DeveloperAPI
@ludwig_dataclass
class BaseAdapterConfig(schema_utils.BaseMarshmallowConfig, ABC):
    type: str

    pretrained_adapter_weights: Optional[str] = schema_utils.String(
        default=None, description="Path to pretrained weights.", allow_none=True
    )

    postprocessor: LoraPostprocessorConfig = LoraPostprocessorConfigField().get_default_field()

    @abstractmethod
    def to_config(self, **kwargs) -> "PeftConfig":
        pass


@DeveloperAPI
@register_adapter(name="lora")
@ludwig_dataclass
class LoraConfig(BaseAdapterConfig):
    type: str = schema_utils.ProtectedString(
        "lora",
        description=LLM_METADATA["adapter"]["lora"]["type"].long_description,
    )

    r: int = schema_utils.PositiveInteger(
        default=8,
        description="Lora attention dimension.",
        parameter_metadata=LLM_METADATA["adapter"]["lora"]["r"],
    )

    alpha: int = schema_utils.PositiveInteger(
        default=16,
        description="The alpha parameter for Lora scaling.",
        parameter_metadata=LLM_METADATA["adapter"]["lora"]["alpha"],
    )

    dropout: float = schema_utils.NonNegativeFloat(
        default=0.05,
        description="The dropout probability for Lora layers.",
        parameter_metadata=LLM_METADATA["adapter"]["lora"]["dropout"],
    )

    # TODO(travis): figure out why calling this `bias` doesn't work
    bias_type: str = schema_utils.StringOptions(
        options=["none", "all", "lora_only"],
        default="none",
        description="Bias type for Lora.",
    )

    def to_config(self, task_type: str = None, **kwargs) -> "PeftConfig":
        from peft import LoraConfig as _LoraConfig

        return _LoraConfig(
            r=self.r,
            lora_alpha=self.alpha,
            lora_dropout=self.dropout,
            bias=self.bias_type,
            task_type=task_type,
        )

    @classmethod
    def name(cls) -> str:
        return "LoRA"

    @classmethod
    def description(cls) -> str:
        return LLM_METADATA["adapter"]["lora"]["type"].long_description


@DeveloperAPI
@ludwig_dataclass
class BasePromptLearningConfig(BaseAdapterConfig):
    """Config for prompt learning adapters. Not meant to be used directly.

    Adapted from https://github.com/huggingface/peft/blob/main/src/peft/utils/config.py (PromptLearningConfig)
    """

    num_virtual_tokens: int = schema_utils.PositiveInteger(
        default=8,
        description="Number of virtual tokens to add to the prompt. Virtual tokens are used to control the behavior of "
        " the model during inference. ",
        parameter_metadata=LLM_METADATA["adapter"]["prompt_learning"]["num_virtual_tokens"],
    )

    token_dim: Optional[int] = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="The hidden embedding dimension of the base transformer model.",
    )

    num_transformer_submodules: Optional[int] = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="The number of transformer submodules in the base transformer model.",
    )

    num_attention_heads: Optional[int] = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="The number of attention heads in the base transformer model.",
    )

    num_layers: Optional[int] = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="The number of layers in the base transformer model.",
    )


# TODO(travis): fix text generation when using prompt tuning:
#     RuntimeError: shape '[-1, 17]' is invalid for input of size 9
# @DeveloperAPI
# @register_adapter("prompt_tuning")
# @ludwig_dataclass
# class PromptTuningConfig(BasePromptLearningConfig):
#     """Adapted from https://github.com/huggingface/peft/blob/main/src/peft/tuners/prompt_tuning.py."""

#     def __post_init__(self):
#         if self.prompt_tuning_init == "TEXT" and not self.prompt_tuning_init_text:
#             raise ConfigValidationError(
#                 "Must provide `prompt_tuning_init_text` when `prompt_tuning_init` is set to `TEXT`."
#             )

"""#     type: str = schema_utils.ProtectedString("prompt_tuning")"""  # Quotes allow mypy to run without syntax errors.

#     prompt_tuning_init: str = schema_utils.StringOptions(
#         ["RANDOM", "TEXT"],
#         default="RANDOM",
#         description="The type of initialization to use for the prompt embedding. ",
#         parameter_metadata=LLM_METADATA["adapter"]["prompt_tuning"]["prompt_tuning_init"],
#     )

#     prompt_tuning_init_text: str = schema_utils.String(
#         default="",
#         description="The text to use to initialize the prompt embedding.",
#         parameter_metadata=LLM_METADATA["adapter"]["prompt_tuning"]["prompt_tuning_init_text"],
#     )

#     def to_config(self, **kwargs) -> "PeftConfig":
#         from peft import PromptTuningConfig as _PromptTuningConfig

#         return _PromptTuningConfig(
#             num_virtual_tokens=self.num_virtual_tokens,
#             token_dim=self.token_dim,
#             num_transformer_submodules=self.num_transformer_submodules,
#             num_attention_heads=self.num_attention_heads,
#             num_layers=self.num_layers,
#             prompt_tuning_init=self.prompt_tuning_init,
#             prompt_tuning_init_text=self.prompt_tuning_init_text,
#             **kwargs
#         )


# TODO(travis): fix prefix tuning and p-tuning to work with DDP
# @DeveloperAPI
# @register_adapter("prefix_tuning")
# @schema_utils.ludwig_dataclass
# class PrefixTuningConfig(BasePromptLearningConfig):
#     """Adapted from https://github.com/huggingface/peft/blob/main/src/peft/tuners/prefix_tuning.py."""

"""#     type: str = schema_utils.ProtectedString("prefix_tuning")"""  # Quotes allow mypy to run without syntax errors.

#     encoder_hidden_size: Optional[int] = schema_utils.Integer(
#         default=None,
#         allow_none=True,
#         description="The hidden embedding dimension of the prompt encoder.",
#     )

#     prefix_projection: bool = schema_utils.Boolean(
#         default=False,
#         description="Whether to use a projection layer in the prompt encoder to project the prefix tokens",
#     )

#     def to_config(self, task_type: str = None, **kwargs) -> "PeftConfig":
#         from peft import PrefixTuningConfig as _PrefixTuningConfig

#         return _PrefixTuningConfig(
#             num_virtual_tokens=self.num_virtual_tokens,
#             token_dim=self.token_dim,
#             num_transformer_submodules=self.num_transformer_submodules,
#             num_attention_heads=self.num_attention_heads,
#             num_layers=self.num_layers,
#             encoder_hidden_size=self.encoder_hidden_size,
#             prefix_projection=self.prefix_projection,
#             task_type=task_type,
#         )


# @DeveloperAPI
# @register_adapter("p_tuning")
# @ludwig_dataclass
# class PTuningConfig(BasePromptLearningConfig):
"""#     type: str = schema_utils.ProtectedString("p_tuning")"""  # Quotes allow mypy to run without syntax errors.

#     encoder_reparameterization_type: str = schema_utils.StringOptions(
#         ["MLP", "LSTM"],
#         default="MLP",
#         allow_none=False,
#         description="The type of reparameterization to use for the prompt encoder.",
#     )

#     encoder_hidden_size: Optional[int] = schema_utils.PositiveInteger(
#         default=None,
#         allow_none=True,
#         description="The hidden embedding dimension of the prompt encoder.",
#     )

#     encoder_num_layers: Optional[int] = schema_utils.PositiveInteger(
#         default=2,
#         allow_none=True,
#         description="The number of layers in the prompt encoder.",
#     )

#     encoder_dropout: Optional[float] = schema_utils.FloatRange(
#         default=0.0,
#         min=0.0,
#         max=1.0,
#         description="The dropout probability for the prompt encoder.",
#     )

#     def to_config(self, task_type: str = None, **kwargs) -> "PeftConfig":
#         from peft import PromptEncoderConfig as _PromptEncoderConfig

#         return _PromptEncoderConfig(
#             num_virtual_tokens=self.num_virtual_tokens,
#             token_dim=self.token_dim,
#             num_transformer_submodules=self.num_transformer_submodules,
#             num_attention_heads=self.num_attention_heads,
#             num_layers=self.num_layers,
#             encoder_reparameterization_type=self.encoder_reparameterization_type,
#             encoder_hidden_size=self.encoder_hidden_size,
#             encoder_num_layers=self.encoder_num_layers,
#             encoder_dropout=self.encoder_dropout,
#             task_type=task_type,
#         )


@DeveloperAPI
@register_adapter("adalora")
@ludwig_dataclass
class AdaloraConfig(LoraConfig):
    """Adapted from https://github.com/huggingface/peft/blob/main/src/peft/tuners/adalora.py."""

    type: str = schema_utils.ProtectedString(
        "adalora",
        description=LLM_METADATA["adapter"]["adalora"]["type"].long_description,
    )

    target_r: int = schema_utils.PositiveInteger(
        default=8,
        description="Target Lora Matrix Dimension. The target average rank of incremental matrix.",
    )

    init_r: int = schema_utils.PositiveInteger(
        default=12,
        description="Initial Lora Matrix Dimension. The initial rank for each incremental matrix.",
    )

    tinit: int = schema_utils.NonNegativeInteger(
        default=0,
        description="The steps of initial fine-tuning warmup.",
    )

    tfinal: int = schema_utils.NonNegativeInteger(
        default=0,
        description="The steps of final fine-tuning warmup.",
    )

    delta_t: int = schema_utils.NonNegativeInteger(
        default=1,
        description="The time internval between two budget allocations. The step interval of rank allocation.",
    )

    beta1: float = schema_utils.FloatRange(
        default=0.85,
        min=0.0,
        max=1.0,
        description="The hyperparameter of EMA for sensitivity smoothing.",
    )

    beta2: float = schema_utils.FloatRange(
        default=0.85,
        min=0.0,
        max=1.0,
        description=" The hyperparameter of EMA for undertainty quantification.",
    )

    orth_reg_weight: float = schema_utils.FloatRange(
        default=0.5,
        min=0.0,
        max=1.0,
        description="The coefficient of orthogonality regularization.",
    )

    total_step: Optional[int] = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="The total training steps that should be specified before training.",
    )

    rank_pattern: Optional[dict] = schema_utils.Dict(
        default=None,
        allow_none=True,
        description="The allocated rank for each weight matrix by RankAllocator.",
    )

    def to_config(self, **kwargs) -> "PeftConfig":
        from peft import AdaLoraConfig as _AdaLoraConfig

        return _AdaLoraConfig(
            r=self.r,
            lora_alpha=self.alpha,
            lora_dropout=self.dropout,
            bias=self.bias_type,
            target_r=self.target_r,
            init_r=self.init_r,
            tinit=self.tinit,
            tfinal=self.tfinal,
            deltaT=self.delta_t,
            beta1=self.beta1,
            beta2=self.beta2,
            orth_reg_weight=self.orth_reg_weight,
            total_step=self.total_step,
            rank_pattern=self.rank_pattern,
        )

    @classmethod
    def name(cls) -> str:
        return "AdaLoRA"

    @classmethod
    def description(cls) -> str:
        return LLM_METADATA["adapter"]["adalora"]["type"].long_description


@DeveloperAPI
@register_adapter("adaption_prompt")
@ludwig_dataclass
class AdaptionPromptConfig(BaseAdapterConfig):
    """Adapted from https://github.com/huggingface/peft/blob/main/src/peft/tuners/adaption_prompt/config.py."""

    def __post_init__(self):
        if not self.adapter_len:
            raise ConfigValidationError(
                "`adapter_len` must be set to a value greater than 0 when finetuning is enabled and the adapter"
                "type is `adaption_prompt`. This is the length of the adaption prompt to insert."
            )

        if not self.adapter_layers:
            raise ConfigValidationError(
                "`adapter_layers` must be set to a value greater than 0 when finetuning is enabled and the adapter"
                "type is `adaption_prompt`. This is the number of adapter layers to insert."
            )

    type: str = schema_utils.ProtectedString(
        "adaption_prompt",
        description=LLM_METADATA["adapter"]["adaption_prompt"]["type"].long_description,
    )

    adapter_len: int = schema_utils.PositiveInteger(
        default=4,
        description="Number of adapter tokens to insert.",
        parameter_metadata=LLM_METADATA["adapter"]["adaption_prompt"]["adapter_len"],
    )

    adapter_layers: int = schema_utils.PositiveInteger(
        default=1,
        allow_none=False,
        description="Number of adapter layers to insert (from the top).",
        parameter_metadata=LLM_METADATA["adapter"]["adaption_prompt"]["adapter_layers"],
    )

    def to_config(self, task_type: str = None, **kwargs) -> "PeftConfig":
        from peft import AdaptionPromptConfig as _AdaptionPromptConfig

        return _AdaptionPromptConfig(
            adapter_len=self.adapter_len,
            adapter_layers=self.adapter_layers,
            task_type=task_type,
        )

    @classmethod
    def name(cls) -> str:
        return "Adaption Prompt"

    @classmethod
    def description(cls) -> str:
        return LLM_METADATA["adapter"]["adaption_prompt"]["type"].long_description


@DeveloperAPI
def get_adapter_conds():
    conds = []
    for adapter_type, adapter_cls in adapter_registry.items():
        other_props = schema_utils.unload_jsonschema_from_marshmallow_class(adapter_cls)["properties"]
        schema_utils.remove_duplicate_fields(other_props)
        preproc_cond = schema_utils.create_cond(
            {"type": adapter_type},
            other_props,
        )
        conds.append(preproc_cond)
    return conds


@DeveloperAPI
def AdapterDataclassField(default: Optional[str] = None):
    description = "Whether to use parameter-efficient fine-tuning"

    class AdapterSelection(schema_utils.TypeSelection):
        def __init__(self):
            super().__init__(
                registry=adapter_registry,
                default_value=default,
                description=description,
                parameter_metadata=None,
                allow_str_value=True,
                allow_none=True,
            )

        def get_schema_from_registry(self, key: str) -> Type[schema_utils.BaseMarshmallowConfig]:
            return adapter_registry[key]

        @staticmethod
        def _jsonschema_type_mapping():
            return {
                "oneOf": [
                    {
                        "type": "object",
                        "properties": {
                            "type": {
                                "type": "string",
                                "enum": list(adapter_registry.keys()),
                                "description": "The type of PEFT adapter to use during fine-tuning",
                            },
                        },
                        "title": "Perform parameter efficient fine-tuning",
                        "allOf": get_adapter_conds(),
                        "required": ["type"],
                        "description": "The type of PEFT adapter to use during fine-tuning",
                        "parameter_metadata": convert_metadata_to_json(LLM_METADATA["adapter"]["_oneOf"]["allOf"]),
                    },
                    {
                        "type": "null",
                        "title": "adapter_null_option",
                        "description": "Disable the adapter.",
                        "parameter_metadata": convert_metadata_to_json(LLM_METADATA["adapter"]["_oneOf"]["none"]),
                    },
                ],
                "title": "adapter_options",
                "description": "Whether to use parameter-efficient fine-tuning",
                "parameter_metadata": convert_metadata_to_json(LLM_METADATA["adapter"]["_meta"]),
                "default": default,
            }

    return AdapterSelection().get_default_field()



# register trainers

import ludwig.trainers.trainer  # noqa: F401

try:
    import ludwig.trainers.trainer_lightgbm  # noqa: F401
except ImportError:
    pass


try:
    import ludwig.trainers.trainer_llm  # noqa: F401
except ImportError:
    pass
